require 'test/unit'
require File.dirname(__FILE__) + "/data_group.rb"
require File.dirname(__FILE__) + "/test_helper.rb"

class DataGroupTest < Test::Unit::TestCase

  include TestHelper

  def setup
    @one_data_point = DataGroup.new([ DataPoint.new(0, 0, 0.5, :yes) ])
    @artificial_data = DataGroup.parse_from_file(ARTIFICIAL_POINTS)
    @data = DataGroup.parse_from_file(SMALL_DATA_FILE)
    @more_data = DataGroup.parse_from_file(NEW_FORMAT_DATA_FILE)
    @more_artificial_data = DataGroup.parse_from_file(LEVEL_TEST_DATA_FILE)
  end

  def test_long_methods
    assert_in_delta -1.1830299622427, @data.log10_prob_fixed_threshold_fixed_epsilon(0.5, 0.1), 1E-7
    assert_in_delta -1.50514997831991, @data.log10_prob_fixed_epsilon_marginal_threshold(0.5), 1E-7
    assert_in_delta -1.39269084298181, @data.log10_prob_marginal_equal_epsilon_marginal_threshold, 1E-7
    assert_in_delta -1.51447202394214, @data.log10_prob_marginal_different_epsilon_marginal_threshold, 1E-7
    assert_in_delta -1.46142626619314, @data.log10_prob_biased_coin, 1E-7
  end

  def test_threshold_walk
    assert_equal [0, 1, 2, 4, 6], @data.thresholds.map { |th| th.index }
    assert_equal [0.040464039814292, 0.0416337474917464 - 0.040464039814292, 0.100339749290499 - 0.0416337474917464, 1 - 0.100339749290499, 0], @data.thresholds.map { |th| th.interval_length }
    assert_equal @data.thresholds, @data.thresholds
    new_thresholds = []
    @data.walk_thresholds { |th| new_thresholds.push(th) }
    assert_equal @data.thresholds, new_thresholds
  end

  def test_counters_at
    assert_equal Counters.new(2, 1, 2, 0), @data.counters_at(0.5)
  end

  def ml_cutoff(data_group, eps_low, eps_high)
    data_group.max_likelihood_threshold(eps_low, eps_high).cutoff
  end

  def map_cutoff(data_group, eps_low, eps_high)
    data_group.best_threshold(eps_low, eps_high).cutoff
  end
    
  def test_best_threshold
    assert_equal 0.35, ml_cutoff(@artificial_data, 0.01, 0.01)
    assert_equal 0.35, map_cutoff(@artificial_data, 0.01, 0.01)

    assert_in_delta 0.07098674839, ml_cutoff(@data, 0.01, 0.01), 1E-10
    assert_in_delta 0.07098674839, map_cutoff(@data, 0.01, 0.01), 1E-10

    assert_in_delta 0.09169690414, ml_cutoff(@more_data, 0.01, 0.01), 1E-10
    assert_in_delta 0.09169690414, map_cutoff(@more_data, 0.01, 0.01), 1E-10

    # Here the prior on the threshold range [~0.1, 1.0] outweighs the
    # preference to account for the positive example at ~0.1 as a
    # correct yes (prob. 0.99) vs an incorrect no (prob. 0.5).
    assert_in_delta 0.09169690414, ml_cutoff(@more_data, 0.5, 0.01), 1E-10
    assert_in_delta 0.55016987464, map_cutoff(@more_data, 0.5, 0.01), 1E-10
  end

  def test_more_best_threshold
    assert_in_delta 0.15, ml_cutoff(@more_artificial_data, 0.01, 0.01), 1E-10
    assert_in_delta 0.15, map_cutoff(@more_artificial_data, 0.01, 0.01), 1E-10

    assert_in_delta 0.35, ml_cutoff(@more_artificial_data, 0.5, 0.01), 1E-10
    assert_in_delta 0.35, map_cutoff(@more_artificial_data, 0.5, 0.01), 1E-10

    # Here the prior on the threshold range [0.6, 1.0] outweighs the
    # preference to account for the three positive 0.6 examples as
    # correct yesses (prob. 0.9) as opposed to incorrect no's (prob. 0.82)
    assert_in_delta 0.55, ml_cutoff(@more_artificial_data, 0.82, 0.01), 1E-10
    assert_in_delta 0.80, map_cutoff(@more_artificial_data, 0.82, 0.01), 1E-10
  end

  def test_trivial_yes_probability
    assert_in_delta 0.5 * 0.9 + 0.1 * 0.1, @one_data_point.yes_probability_fixed_epsilons(0.25, 0.1, 0.1), 1E-15
    assert_in_delta 0.9 * 0.9 + 0.5 * 0.1, @one_data_point.yes_probability_fixed_epsilons(0.75, 0.1, 0.1), 1E-15
  end

  def test_yes_probability
    assert_in_delta 0.5, @artificial_data.yes_probability_fixed_epsilons(0.5, 0.5, 0.5), 1E-15
    assert_in_delta 0.1, @artificial_data.yes_probability_fixed_epsilons(0.001, 0.1, 0.1), 1E-3
    assert_in_delta 1 - 0.1, @artificial_data.yes_probability_fixed_epsilons(0.999, 0.1, 0.1), 1E-3
    assert_in_delta 0.80335, @artificial_data.yes_probability_fixed_epsilons(0.6, 0.1, 0.1), 1E-4
  end

  def test_equality
    assert_equal(@data, DataGroup.parse_from_file(SMALL_DATA_FILE))
  end

end
