require File.dirname(__FILE__) + "/general_utilities/general_utilities.rb"
require File.dirname(__FILE__) + "/probabilities/probabilities.rb"
require File.dirname(__FILE__) + "/data_group.rb"
require File.dirname(__FILE__) + "/parsing/parseable.rb"
require File.dirname(__FILE__) + "/parsing/judgement_request.rb"
require File.dirname(__FILE__) + "/trec_f_measure.rb"

# A training data set, consisting of DataPoints grouped into
# DataGroups.  A DataSet embodies the model that epsilons are global
# to the whole data set, but thresholds are per DataGroup.  Modeling
# epsilons that are local to components of one's whole collection of
# data (e.g. questions) can be done by maintaining multiple DataSet
# objects.
class DataSet

  extend Parseable

  attr_accessor :verbosity

  def initialize(data_group_hash, &point_hasher)
    @data_group_hash = data_group_hash
    @point_hasher = point_hasher
    @verbosity = :normal
  end

  # TODO: Figure out how to define data set equality.  The problem is that
  # equality on procedures presents problems (see test).
  def == (other)
    (@data_group_hash == other.instance_eval { @data_group_hash }) and
      (@point_hasher == other.instance_eval { @point_hasher })
  end

  # Parses DataPoints from the given strings, groups them according to
  # the allocation method (see #allocate), and returns the resulting
  # DataSet.
  def self.parse_from_strings(strings, parse_type = :block, &block)
    self.allocate(DataPoint.parse_from_strings(strings), parse_type, &block)
  end

  def reallocate(allocation_type = :block, &block)
    DataSet.allocate(all_data_points, allocation_type, &block)
  end

  def self.allocate(data_points, allocation_type = :block)
    case allocation_type
    when :block
      DataSet.allocate_with_block(data_points) { |pt| yield pt }
    when :one_group
      DataSet.allocate(data_points) { |pt| 1 }
    when :question
      DataSet.allocate(data_points) { |pt| pt.question_id }
    when :nugget
      DataSet.allocate(data_points) { |pt| pt.question_id + "," + pt.nugget_id }
    else
      raise "Unknown allocation type."
    end
  end

  # Groups all the given data points into DataGroups as keyed by the
  # values returned by the supplied block and returns the resulting
  # DataSet.
  def self.allocate_with_block(data_points, &block)
    point_groups = data_points.inject(Hash.new { || [] }) { |hash, point| key = yield(point); hash[key] = hash[key].push(point); hash }
    data_group_hash = Hash.new
    point_groups.each_pair do |key, points|
      data_group_hash[key] = DataGroup.new(points)
    end
    DataSet.new(data_group_hash, &block)
  end


  def data_groups
    @data_group_hash.values
  end

  def all_data_points
    data_groups.map { |gr| gr.data_points }.flatten
  end

  def only_group
    if data_groups.length == 1
      data_groups[0]
    else
      raise "Requested the only group of a multigroup data set."
    end
  end

  def total_points
    data_groups.inject(0) { |sum, group| sum + group.data_points.length }
  end


  def yes_probability_fixed_epsilons(data_point, eps_low, eps_high)
    @data_group_hash[@point_hasher.call(data_point)].yes_probability_fixed_epsilons(data_point.score, eps_low, eps_high)
  end

  # The probability that the given new data point should be
  # judged "yes" given the training data (and grouping method) 
  # embodied by this data set.
  # Returns 0 for points outside any group.
  def yes_probability(data_point)
    group = @data_group_hash[@point_hasher.call(data_point)]
    if group
      group.yes_probability_with_distribution(data_point.score, epsilon_posterior)
    else
      0
    end
  end

  instrument :yes_probability


  def log10_prob_separate_different_epsilon_and_threshold
    data_groups.inject(0) { |sum, group| sum + group.log10_prob_marginal_different_epsilon_marginal_threshold }
  end

  def log10_prob_separate_equal_epsilon_and_threshold
    data_groups.inject(0) { |sum, group| sum + group.log10_prob_marginal_equal_epsilon_marginal_threshold }
  end

  def log10_prob_fixed_epsilon_marginal_thresholds(eps_low, eps_high = eps_low)
    data_groups.inject(0) { |sum, group| sum + group.log10_prob_fixed_epsilon_marginal_threshold(eps_low, eps_high) }
  end

  def plot_MAP_thresholds(eps_low, eps_high = eps_low)
    puts "#{group.question_id}\t#{group.nugget_id}\t#{group.best_threshold(eps_low, eps_high)}"
  end

  def log10_lin_lin_plot_likelihood_marginal_threshold_vs_epsilon
    0.1.step(0.9, 0.1) do |eps_low|
      STDERR.puts eps_low if @verbosity == :normal || @verbosity == :loud
      0.1.step(0.9, 0.1) do |eps_high|
        puts "#{eps_low} #{eps_high} #{log10_prob_fixed_epsilon_marginal_thresholds(eps_low, eps_high)}"
      end
    end

    0.01.step(0.2, 0.01) do |eps_low|
      STDERR.puts eps_low if @verbosity == :normal || @verbosity == :loud
      0.01.step(0.35, 0.01) do |eps_high|
        puts "#{eps_low} #{eps_high} #{log10_prob_fixed_epsilon_marginal_thresholds(eps_low, eps_high)}"
      end
    end
  end

  # Global different epsilons
  def best_epsilon
    ans = unnormalized_posterior_log10.max { |ep_set1, ep_set2| ep_set1[2] <=> ep_set2[2] }
    STDERR.puts "Best epsilons: low #{ans[0]}, high #{ans[1]}" if @verbosity == :normal || @verbosity == :loud
    ans
  end

  # Global different epsilons
  def unnormalized_posterior_log10
    STDERR.puts "Computing unnormalized posterior" if @verbosity == :normal || @verbosity == :loud
    ans = []
    eps_low_step = 0.01
    eps_low_num = 20
    eps_high_step = 0.01
    eps_high_num = 35
    eps_prior = Math.log10(1.0/eps_low_num) + Math.log10(1.0/eps_high_num)
    1.upto(eps_low_num) do |eps_low_count|
      eps_low = eps_low_count * eps_low_step # Floating roundoff error gymnastics
      STDERR.puts "eps_low: #{eps_low}" if @verbosity == :normal || @verbosity == :loud
      1.upto(eps_high_num) do |eps_high_count|
        eps_high = eps_high_count * eps_high_step
        STDERR.puts "eps_high: #{eps_high}" if @verbosity == :loud
        cur = [eps_low, eps_high, log10_prob_fixed_epsilon_marginal_thresholds(eps_low, eps_high) + eps_prior]
        ans = ans + [cur]
      end
    end
    ans
  end

  memoize :unnormalized_posterior_log10

  def log10_prob_global_different_epsilon_local_threshold
    Arithmetic.logspace_sum(unnormalized_posterior_log10.map { |ep_set| ep_set[2] } )
  end

  # Global different epsilons
  def epsilon_posterior
    unnorm_post = unnormalized_posterior_log10
    normalization_constant_log10 = Arithmetic.logspace_sum(unnorm_post.map { |ep_set| ep_set[2] } )
    DiscreteProbabilityDistribution.new(unnorm_post.map { |ep_set| [[ep_set[0], ep_set[1]], 10.0**(ep_set[2] - normalization_constant_log10)] })
  end

  instrument :epsilon_posterior

  memoize :epsilon_posterior


  # Per-question expected f-measure for one system.
  # judgement_requests must be complete and exact -- those and only those from this question
  # nugget_vitality is a hash from nids to true/false
  #   indications of whether the nugget is vital
  def expected_f(judgement_requests, nugget_vitality, beta = 3)
    length = judgement_requests.map { |req| req.length }.inject(0) { |sum, len| sum + len }
    num_vital = nugget_vitality.values.select { |vitality| vitality }.length
    data_points = judgement_requests.map { |req| req.data_points }.flatten
    ok_vital_count_distribution(data_points, nugget_vitality).integrate do |ok_ct, vital_ct|
      TrecFMeasure.new(:okay_count => ok_ct, :vital_count => vital_ct, :num_vital => num_vital, :length => length, :beta => beta)
    end
  end

  def ok_vital_count_distribution(data_points, nugget_vitality)
    vital_nuggets = nugget_vitality.keys.select { |key| nugget_vitality[key] }
    vital_count_distribution = count_distribution(vital_nuggets.map { |nugget| nugget_probability(nugget, data_points) })
    okay_nuggets = nugget_vitality.keys.select { |key| !nugget_vitality[key] }
    okay_count_distribution = count_distribution(okay_nuggets.map { |nugget| nugget_probability(nugget, data_points) })
    okay_count_distribution.times vital_count_distribution
  end

  def count_distribution(one_probabilities)
    if one_probabilities.empty?
      DiscreteProbabilityDistribution.singleton 0
    else
      rest_distribution = count_distribution(one_probabilities[1..-1])
      DiscreteProbabilityDistribution.new([ [ rest_distribution, 1 - one_probabilities[0] ],
                                            [ rest_distribution.map { |count| count + 1 }, one_probabilities[0] ] ]).flatten_once
    end
  end

  # The probability with which the given nugget should be judged as
  # contained in the given data points (based on their scores for it).
  # Assumes that all given data points refer to the same question as
  # the given nugget id, and silently ignores those that do not refer
  # to the given nugget.
  def nugget_probability(nugget, data_points)
    1 - data_points.select { |pt| pt.nugget_id == nugget }.inject(1) { |prod, pt| prod * (1 - yes_probability(pt)) }
  end

  instrument :nugget_probability

  # A hash { nid => prob } of probabilities that the response
  # requested by this judgement request contains any given nugget.
  # The optional second argument specifies the key set of the returned
  # hash.  By default, this is all the nuggets that appear in the
  # given request.
  def nugget_probabilities(judge_request, nuggets = judge_request.nuggets)
    nuggets.inject(Hash.new(0)) { |hash, nid| hash[nid] = yes_probability(judge_request.data_point(nid)); hash }
  end

end
