require File.dirname(__FILE__) + "/threshold.rb"

# A collection of the maximum likelihood thresholds for a data group
# or data set for every error rate region.
#
# The way this class works and the reasons why it works are detailed
# in Nuggeteer/doc/math/error_rate_space.{tex|dvi|pdf}.
class ErrorRateSpace

  attr_reader :list

  def initialize(data_group)
    # This is directly discussed in the paper, in section 3.2.
    @list = [ ErrorClassBoundary.new(0, :threshold => data_group.lowest_threshold) ]
    data_group.walk_thresholds do |th|
      next if th.interval_length <= 0
      add_threshold!(th)
    end
    freeze
  end

  # The maximum likelihood threshold for this data and the given error rates.
  def max_likelihood_threshold(eps_low, eps_high = eps_low)
    # This ErrorClassQuery thing is rather hacked.  See the
    # class docs below.
    @list.binary_search(ErrorClassQuery.new(eps_low, eps_high)).threshold
  end

  private
  # This is directly discussed in the paper, in section 3.2.
  def add_threshold!(threshold)
    while @list.length > 0 && @list[-1].step_acceptable?(threshold.counters) do
      @list.pop
    end
    if @list.length == 0
      @list.push ErrorClassBoundary.new(0, :threshold => threshold)
    else
      @list.push ErrorClassBoundary.new(@list[-1].step_mistake_ratio(threshold.counters), :threshold => threshold)
    end
  end

  # A rather unfortunate hack to use binary search over a list
  # of ErrorClassBoundary objects without knowing which number any
  # given pair of epsilons corresponds to.
  # This relies on binary search only asking 'query <=> value'
  # questions, and not 'value <=> query'
  class ErrorClassQuery

    include Comparable

    def initialize(eps_low, eps_high)
      @mistake_tolerance = eps_low / (1.0 - eps_high)
      @judgement_cutoff = eps_high / (1.0 - eps_low)
    end

    def <=>(error_class)
      return -1 if error_class.mistake_ratio >= 1.0 - 1e-15
      wrong_to_right_ratio = error_class.mistake_ratio / (1.0 - error_class.mistake_ratio)
      if @mistake_tolerance ** wrong_to_right_ratio >= @judgement_cutoff
        1
      else
        -1
      end
    end

  end

end

# A boundary in the error rate space.  Knows the characteristics of
# the maximum likelihood threshold(s) for error rates just beyond the
# boundary (namely those that accept this mistake ratio and no greater
# ones).
class ErrorClassBoundary

  attr_reader :mistake_ratio, :threshold, :counters

  def initialize(mistake_ratio, payload = {})
    @mistake_ratio = mistake_ratio
    if payload.has_key? :threshold
      @threshold = payload[:threshold]
      @counters = @threshold.counters
    else
      @threshold = nil
      @counters = payload[:counters]
    end
  end

  def step_mistake_ratio(new_counters)
    @counters.new_mistake_ratio(new_counters)
  end

  def step_acceptable?(new_counters)
    @mistake_ratio >= step_mistake_ratio(new_counters)
  end

  def ==(other)
    other.kind_of?(ErrorClassBoundary) &&
      @mistake_ratio == other.mistake_ratio &&
      @counters == other.counters &&
      @threshold == other.threshold
  end

end
