class Array

  # Returns the greatest element of self less than the argument, or
  # nil if the argument is not greater than any element of self.
  # Compares using the given block, if any, or the <=> operator.  self
  # *must* already be sorted consistently with the comparison function
  # being used, or the results are unspecified.
  #
  # The following options are supported:
  # :return determines whether to return the actual
  #   element found or its index.  Supported values are :index, :object (default),
  #   :index_and_object, and :object_and_index.  The latter return pairs.
  # :find determines which object, in relation to the key, to find.
  #   Supported values are :next_smaller (default) and :next_larger
  def binary_search(key, options = {}, &block)
    return binary_search(key, options) { |key, obj| key <=> obj } unless block_given?
    if self.length == 0
      nil
    else
      if options[:find] == :next_larger
        if yield(key, self[-1]) >= 0
          nil
        else
          binary_search_return(options[:return],
                               do_binary_search_up(key, -1, self.length - 1, &block))
        end
      else
        if yield(key, self[0]) <= 0
          nil
        else
          binary_search_return(options[:return],
                               do_binary_search_down(key, 0, self.length, &block))
        end
      end
    end
  end

  private
  # Recursion invariant: self[min_ind] < key <= self[max_ind]
  def do_binary_search_down(key, min_ind, max_ind, &block)
    mid_ind = (min_ind + max_ind)/2
    if mid_ind == min_ind
      min_ind
    else
      if yield(key, self[mid_ind]) > 0
        do_binary_search_down(key, mid_ind, max_ind, &block)
      else
        do_binary_search_down(key, min_ind, mid_ind, &block)
      end
    end
  end

  # Recursion invariant: self[min_ind] <= key < self[max_ind]
  def do_binary_search_up(key, min_ind, max_ind, &block)
    mid_ind = (min_ind + max_ind + 1)/2
    if mid_ind == max_ind
      max_ind
    else
      if yield(key, self[mid_ind]) < 0
        do_binary_search_up(key, min_ind, mid_ind, &block)
      else
        do_binary_search_up(key, mid_ind, max_ind, &block)
      end
    end
  end

  def binary_search_return(type, index)
    case type
    when :index: index
    when :object: self[index]
    when :index_and_object: [index, self[index]]
    when :object_and_index: [self[index], index]
    else
      self[index]
    end
  end

end

# Thoughts about binary search:
# Should offer comparison by block.
# Probably should document comparison asymmetry
# Can offer options like 
#  :return => :greatest_lesser | :least_greater | :both
#  :strict_comparison => :yes | :no
# Or, perhaps, offer a five-way return option, to the
# tune of
#  :return => :lt | :le | :eq | :ge | :gt,
# and return the triple [ :lt, :eq, :gt ] (the middle of which
# is expected to be nil) by default.
