class Module

  def memoize(name)
    computer_name = ("compute_" + name.to_s).to_sym
    alias_method computer_name, name
    define_method(name) do |*args|
      @memoize_caches ||= {}
      @memoize_caches[name] ||= {}
      @memoize_caches[name][args] ||= self.send computer_name, *args
    end
  end

end

# Considerations on exporting this memoize function to the rest
# of the world:
#
# Documentation.  Explain carefully that it is best used on
# expensive computations over immutable objects.  What are some
# good examples of such?
#
# Functionality.  Some support can be added for handling semi-mutable
# objects (those that can mutate, but at sufficiently rare times,
# predictable by the programmer).  Some such things:
# 1) A 'clear cache' method.  This would be for objects that want
# memoization most of the time, but mutate occasionally, so need the
# cache cleared from time to time.
# 2) Per-object memoization.  If the class doesn't know whether 
# the objects will be immutable, but individual objects do, then
# they can orchestrate their own memoization.  As a specific
# case, perhaps a memoize_on_freeze option?  How, exactly,
# does one implement per-object memoization?  Will it cause
# appreciable problems with overabundant anonymous classes?
# 3) An 'unmemoize' method.  This would be for objects that
# mutate a frequently in general, but are predictably immutable during
# certian periods of intense computation, and would like to benefit
# from memoization then.  One slightly poor example of this is
# computing the standard deviation of an array like this:
# def std_dev
#   Math.sqrt(inject(0) { |sum, obj| sum + (obj^2 - average^2).abs } / length)
# end
# Such an implemetation runs in quadratic time unless the average
# method is memoized.  One may not wish to memoize the average
# method in general, but maybe on this particular array at this
# particular time.  The example is not a very good one because
# in this case it is too easy to add a local variable and modify
# the inject call to use it, but in a more complex situation that
# may not be possible.  Warning: This feature smacks of poor
# interaction with threading.
#
# Interface.  Memoizing multiple things at once, options to memoize to
# control what sorts of things it does, etc.  What interface to offer
# to the other methods?  Do I make one method called clear_cache that
# takes a method name as an argument, or do I generate a
# clear_foo_cache method every time I call memoize :foo?  Should this
# be done for immutable objects?  What should the default be?
# Should there be a clear_all_caches?  Should calling clear_cache 
# with no arguments do that?
#
# Presentation.  Should this be added to the core Module class?
# On the one hand, it belongs there, but on the other hand,
# it feels one level higher than the rest of the methods of Module.
# Should it be in a separate module in the core or standard libraries?
# Should it be included/extended by Module by default?  Should it
# be a gem?  If it's a module in its own right, should it be
# aimed specifically at memoization, or try to live alongside other
# ideas of this kind?  What should that module be called?
