require File.dirname(__FILE__) + "/regex_parseable.rb"
require File.dirname(__FILE__) + "/assignment.rb"
require File.dirname(__FILE__) + "/response.rb"
require File.dirname(__FILE__) + "/debug_line.rb"
require File.dirname(File.dirname(__FILE__)) + "/general_utilities/module.rb"
require File.dirname(File.dirname(__FILE__)) + "/trec_f_measure.rb"

# A collection of a system's responses to Trec other or ciqa
# questions, and how they were judged.  One of these objects is not
# necessarily a complete collection, as it could be responses to
# particular questions, responses that also appear in another run,
# etc.
class ResponseSet

  include Enumerable

  def initialize(options)
    @response_hash = Hash.new([])
    @response_array = (options[:responses] || []).dup
    compute_response_hash!
    parse_judgements!(options[:judgement_file] || options[:file])
    parse_nuggeteer_output!(options[:nuggeteer_file] || options[:file])
  end

  def each(&block)
    @response_array.each(&block)
  end

  def +(other)
    ResponseSet.new(:responses => @response_array +
                    other.instance_eval { @response_array })
  end

  def -(other)
    ResponseSet.new(:responses => @response_array -
                    other.instance_eval { @response_array })
  end

  def ==(other)
    @response_array == other.instance_eval { @response_array }
  end

  alias eql? ==

  def hash
    @response_array.hash
  end

  def length
    @response_array.length
  end

  def qid_filter(question_id)
    case question_id
    when :all: self
    when Array: ResponseSet.new(:responses => question_id.map { |qid| @response_hash[qid] }.flatten)
    else ResponseSet.new(:responses => @response_hash[question_id])
    end
  end

  def response_by_id(question_id, response_id)
    ans = @response_hash[question_id].find { |response| response.response_id == response_id }
    if ans.nil?
      raise "Response not found to question #{question_id} with id '#{response_id}'."
    end
    ans
  end

  def response_by_string(question_id, string)
    ans = @response_hash[question_id].find { |response| response.string == string.downcase.gsub(/\s+/, " ") }
    if ans.nil?
      raise "Response not found to question #{question_id} with string '#{string}'."
    end
    ans
  end

  def identical_response(given_resp)
    @response_hash[given_resp.question_id].find do |r|
      r.string == given_resp.string
    end
  end

  def same_nuggets_assigned?(given_resp)
    my_version = identical_response(given_resp)
    if my_version
      my_version.nuggets_assigned == given_resp.nuggets_assigned
    else
      false
    end
  end

  def qids
    @response_hash.keys.sort
  end

  def common_responses(other_run, question_id = :all)
    ResponseSet.new(:responses => qid_filter(question_id).select { |r| other_run.identical_response r })
  end

  def novel_responses(other_run, question_id = :all)
    ResponseSet.new(:responses => qid_filter(question_id).select { |r| !other_run.identical_response(r) })
  end

  def with_nuggets
    ResponseSet.new(:responses => select { |r| r.any_nugget_assigned? })
  end

  def without_nuggets
    ResponseSet.new(:responses => select { |r| !r.any_nugget_assigned? })
  end

  def identically_judged(other_run)
    ResponseSet.new(:responses => select { |r| other_run.same_nuggets_assigned?(r) })
  end

  def nuggets_assigned(question_id = :all)
    qid_filter(question_id).map { |r| r.nuggets_assigned }.flatten.sort
  end

  def nuggets_lost(other_run, options = {})
    question_id = options[:question] || options[:questions] || :all
    semantics = options[:semantics] || :reinterpretation
    # This relies on Response#nuggets_assigned returning nuggets 
    # as strings that include the question id.
    if semantics == :whole_run
      nuggets_assigned(question_id) - other_run.nuggets_assigned(question_id)
    else
      qid_filter(question_id).common_responses(other_run).nuggets_assigned -
        other_run.qid_filter(question_id).common_responses(self).nuggets_assigned
    end
  end

  def nuggets_gained(other_run, options = {})
    other_run.nuggets_lost(self, options)
  end

  memoize :common_responses
  memoize :novel_responses
  memoize :nuggets_assigned
  memoize :nuggets_lost
  memoize :nuggets_gained

  def nugget_present?(qid, nid)
    @response_hash[qid].inject(false) { |run, resp| run || resp.judgement(nid) == :yes }
  end

  def f_measure_talk(other_run)
    num_nug = common_responses(other_run).nuggets_assigned.length
    num_nug_lost = common_responses(other_run).nuggets_lost(other_run).length
    num_nug_common = num_nug - num_nug_lost
    recall = 1.0 * num_nug_common / num_nug
    precision = 1.0 * num_nug_common / other_run.common_responses(self).nuggets_assigned.length
    f_measure = 2 * precision * recall / (precision + recall)
    printf("%d %d %4.2f %4.2f %4.2f\n", common_responses(other_run).length, num_nug_common, precision, recall, f_measure)
    :done
  end

  def print_data_points
    @response_hash.each { |key, value| value.each { |response| response.print_data_points } }
  end

  # TODO Ill tested.  May not be right.
  def trec_f_measure(qid, nugget_data, beta = 3)
    relevant_responses = qid_filter(qid)
    length = relevant_responses.inject(0) { |sum, r| sum + r.trec_length }
    total_num_vital = nugget_data.num_vital(qid)
    num_vital_present = nugget_data.vital_nuggets(qid).select { |nid| relevant_responses.nugget_present?(qid, nid) }.length
    num_okay_present = nugget_data.okay_nuggets(qid).select { |nid| relevant_responses.nugget_present?(qid, nid) }.length
    TrecFMeasure.new(:okay_count => num_okay_present, :vital_count => num_vital_present, :num_vital => total_num_vital, :length => length, :beta => beta)
  end

  # TODO Ill tested.  May not be right.
  def avg_trec_f_measure(nugget_data, beta = 3)
    qids.inject(0) { |run, qid| run + trec_f_measure(qid, nugget_data, beta).f_measure } / qids.length
  end

  private
  def compute_response_hash!
    @response_array.each { |r| add_response_to_hash! r }
  end

  def parse_judgements!(filename)
    return unless filename
    File.open(filename) do |f|
      f.each do |line|
        if Response.line_matches?(line)
          parse_new_response!(line)
        end
        if Assignment.line_matches?(line)
          parse_new_assignment!(line)
        end
      end
    end
    self
  end

  def parse_new_response!(line)
    add_response! Response.parse_line(line)
  end

  def add_response!(response)
    @response_array.push response
    add_response_to_hash! response
  end

  def add_response_to_hash!(response)
    @response_hash[response.question_id] = @response_hash[response.question_id] + [response]
  end

  def parse_new_assignment!(line)
    assignment = Assignment.parse_line(line)
    response_by_id(assignment.question_id, assignment.response_id).assign_nugget!(assignment.nugget_id)
    @response_hash[assignment.question_id].each { |response| response.nugget_has_been_assigned!(assignment.nugget_id) }
  end

  def parse_nuggeteer_output!(filename)
    return unless filename
    File.open(filename) do |f| 
      f.each do |line| 
        if DebugLine.line_matches?(line)
          parse_new_score!(line)
        end 
      end 
    end
    self
  end

  def parse_new_score!(line)
    add_score!(DebugLine.parse_line(line))
  end

  def add_score!(debug_line)
    response_by_string(debug_line.question_id, debug_line.response_str).add_score!(debug_line.nugget_id, debug_line.score)
  end

end
