require File.dirname(__FILE__) + "/../general_utilities/general_utilities.rb"

class DiscreteProbabilityDistribution

  class InvariantUnsatisfied < StandardError 
  end

  def initialize(object_probability_array)
    @probabilities = Hash.new(0)
    object_probability_array.each do |object, probability|
      @probabilities[object] += probability
    end
    @probabilities.freeze
    check_invariants
  end

  private
  def check_invariants
    raise InvariantUnsatisfied, "Object probability mapping is empty." unless @probabilities.length > 0
    @probabilities.each do |obj, prob|
      raise InvariantUnsatisfied, "Bad probability" unless prob >= 0 && prob <= 1
    end
    raise InvariantUnsatisfied, "Sum of probabilities is not one." unless (1.0 - integrate { |obj| 1 }).abs < 1E-10
  end
  
  public

  attr_reader :probabilities

  # The objects in this distribution, with their probabilites, as an
  # array in order of decreasing probability.
  def descending_probabilities
    @probabilities.sort { |kv1, kv2| kv2[1] <=> kv1[1] }
  end

  memoize :descending_probabilities

  def sample
    value = rand
    @probabilities.each do |obj, prob|
      if value < prob
        return obj
      else
        value -= prob
      end
    end
    raise "Failed to sample a value, tried to reach #{value}."
  end

  # Integrates the given block over this entire probability distribution.
  def integrate
    @probabilities.inject(0) { |sum, obj_prob| sum + yield(obj_prob[0]) * obj_prob[1] }
  end

  # Integrates the given block over this probability distribution,
  # opportunistically ignoring irrelevant objects.  Objects are
  # irrelevant if their probability is small enough that even if they
  # supplied the maximum possible contribution to the sum (given by
  # the upper_bound argument) they would not contribute a change to
  # the net result greater than the given relative error tolerance.
  def integrate_by_relevance(upper_bound, error_tolerance = 1e-15)
    # I run the summation until the number of objects left in the
    # distribution, multiplied by the maximum possible probability
    # mass of any one of them, ceases being significant.  This is
    # wasteful compared to tracking the total amount of remaining
    # probability mass, but the latter should not be done by starting
    # with 1 and subtracting off the probabilities of the objects
    # already processed, because doing so is extremely vulnerable to
    # roundoff error.
    objects_left = descending_probabilities.select { |obj, prob| prob > 0 }.length
    descending_probabilities.inject(0) do |sum, obj_prob|
      if sum * error_tolerance > upper_bound * obj_prob[1] * objects_left
        return sum
      else
        objects_left -= 1
        sum + yield(obj_prob[0]) * obj_prob[1]
      end
    end
  end

  def map
    DiscreteProbabilityDistribution.new(@probabilities.map { |obj, prob| [yield(obj), prob] })
  end

  def flatten_once
    answer = Hash.new(0)
    @probabilities.each do |subdist, probability|
      subdist.probabilities.each do |object, subprob|
        answer[object] += subprob * probability
      end
    end
    DiscreteProbabilityDistribution.new(answer)
  end

  def self.singleton(object)
    DiscreteProbabilityDistribution.new( { object => 1.0 } )
  end

  def self.uniform(objects)
    uniform_probability = 1.0 / objects.length
    DiscreteProbabilityDistribution.new( objects.map { |obj| [obj, uniform_probability] } )
  end

  def self.unnormalized(object_probability_array)
    normalization_constant = 0.0
    object_probability_array.each do |obj, prob|
      normalization_constant += prob
    end
    DiscreteProbabilityDistribution.new object_probability_array.map { |obj, prob| [obj, prob / normalization_constant] }
  end

  def times(other)
    if block_given?
      answer = Hash.new(0)
      @probabilities.each do |obj1, prob1|
        other.probabilities.each do |obj2, prob2|
          answer[yield(obj1, obj2)] += prob1 * prob2
        end
      end
      DiscreteProbabilityDistribution.new(answer)
    else
      times(other) { |obj1, obj2| [obj1, obj2] }
    end
  end

  def power(number, &block)
    if number < 1
      raise StandardError, "I don't know how to raise probability distributions to powers like #{number}."
    end
    if number == 1
      self
    elsif number % 2 == 1
      times(power(number - 1, &block), &block)
    else
      halfpow = power(number / 2, &block)
      halfpow.times(halfpow, &block)
    end
  end

  def truth_probability
    integrate { |obj| if yield(obj): 1 else 0 end }
  end

  def ==(other)
    @probabilities == other.probabilities
  end
end
