require 'test/unit'
require File.dirname(__FILE__) + "/discrete_probability_distribution.rb"
require File.dirname(File.dirname(__FILE__)) + "/test_helper.rb"

class DiscreteProbabilityDistributionTest < Test::Unit::TestCase

  include TestHelper

  def setup
    @fair_coin = DiscreteProbabilityDistribution.new [ [:heads, 0.5], [:tails, 0.5] ]
    @biased_coin = DiscreteProbabilityDistribution.new [ [:tails, 0.2], [:heads, 0.8] ]
    @fair_die = DiscreteProbabilityDistribution.uniform([1, 2, 3, 4, 5, 6])
    @loaded_die = DiscreteProbabilityDistribution.new [ [1, 0.4], [2, 0.3], [3, 0.3] ]
  end

  def test_smoke
    assert_equal({ 13 => 1 }, DiscreteProbabilityDistribution.singleton(13).probabilities)
    assert_equal [ [:heads, 0.8], [:tails, 0.2] ], @biased_coin.descending_probabilities
  end

  def test_normalization
    assert_close({ :heads => 0.8, :tails => 0.2 },
                 DiscreteProbabilityDistribution.unnormalized({:heads => 4, :tails => 1}).probabilities)
  end

  def test_map
    assert_equal({ :foo => 0.5, :bar => 0.5 }, @fair_coin.map { |side| side == :heads ? :foo : :bar }.probabilities)
  end

  def test_integrate
    assert_equal 0.5, @fair_coin.integrate { |side| side == :heads ? 1 : 0 }
    assert_equal(0.5, @fair_coin.integrate_by_relevance(1) do |side|
                   side == :heads ? 1 : 0
                 end)
    assert_equal 1.9, @loaded_die.integrate { |num| num }
    assert_equal 1.9, @loaded_die.integrate_by_relevance(6) { |num| num }
  end

  def test_product
    assert_equal({ [:heads, 1] => 0.2, [:heads, 2] => 0.15, [:heads, 3] => 0.15,
                   [:tails, 1] => 0.2, [:tails, 2] => 0.15, [:tails, 3] => 0.15 },
                 (@fair_coin.times @loaded_die).probabilities)
  end

  def test_flatten_once
    assert_equal({ :heads => 0.65, :tails => 0.35 }, DiscreteProbabilityDistribution.new([ [@fair_coin, 0.5], [@biased_coin, 0.5] ]).flatten_once.probabilities)
  end

  def test_uniform
    assert_equal({ 1 => 0.25, 2 => 0.25, 3 => 0.25, 4 => 0.25 },
                 DiscreteProbabilityDistribution.uniform( [1, 2, 3, 4] ).probabilities)
  end

  def test_power
    assert_equal({ 0 => 1.0/16, 1 => 4.0/16, 2 => 6.0/16, 3 => 4.0/16, 4 => 1.0/16},
                 DiscreteProbabilityDistribution.uniform([0, 1]).power(4) { |a, b| a + b }.probabilities)
    assert_equal({ 0 => 1.0/32, 1 => 5.0/32, 2 => 10.0/32, 3 => 10.0/32, 4 => 5.0/32, 5 => 1.0/32},
                 DiscreteProbabilityDistribution.uniform([0, 1]).power(5) { |a, b| a + b }.probabilities)
  end

  def test_truth_probability
    assert_equal(21.0/36, @fair_die.power(2) { |a, b| a + b }.truth_probability { |roll| roll >= 7 })
    assert_equal(0, @fair_die.power(3) { |a, b| a + b }.truth_probability { |roll| roll < 3 || roll > 18 })
    assert_in_delta(8.0/36, @fair_die.power(2) { |a, b| a + b }.truth_probability { |roll| roll == 7 || roll == 11 }, 1E-10)
  end

  def test_coin_flipping
    n = 500
    coin_flips = (1..2*n).map { @fair_coin.sample }
    num_heads = coin_flips.select { |f| f == :heads }.length
    num_tails = coin_flips.select { |f| f == :tails }.length
    assert_in_delta n, num_heads, n / 10
    assert_in_delta n, num_tails, n / 10
    assert_equal 2*n, num_heads + num_tails
  end

  def test_die_rolling
    n = 500
    die_rolls = (1..6*n).map { @fair_die.sample }
    total = 0
    (1..6).each do |num|
      num_rolled = die_rolls.select { |r| r == num }.length
      assert_in_delta n, num_rolled, n / 10
      total += num_rolled
    end
    assert_equal 6*n, total    
  end
end
