
# A bundle of the numbers relevant to a TREC F-measure computation for
# a system-question pair, including the beta, the length of the given
# answer(s), the total number of vital nuggets, the number of okay and
# vital nuggets in the answer(s), and the allowance, precision,
# recall, and F-measure, as specified in the TREC documentation.
#
# An instance of TrecFMeasure can also represent a bundle of
# expectations of these numbers, or a partial sum during a computation
# of such a bundle of expectations.  Note that since expectations do
# not divide, the expected F-measure need not equal the F-measure
# derived from the expected precision and the expected recall.
class TrecFMeasure

  class IncompatibleGlobalDataError < StandardError
  end

  INCOMPUTABLE_FIELDS = [:okay_count, :vital_count]
  QUESTION_GLOBAL_FIELDS = [:num_vital, :length, :beta]
  COMPUTABLE_FIELDS = [:allowance, :precision, :recall, :f_measure]

  LOCAL_FIELDS = INCOMPUTABLE_FIELDS + COMPUTABLE_FIELDS
  ALL_FIELDS = INCOMPUTABLE_FIELDS + QUESTION_GLOBAL_FIELDS + COMPUTABLE_FIELDS

  ALL_FIELDS.each do |field|
    attr_reader field
  end

  def field(symbol)
    if ALL_FIELDS.find { |elt| elt == symbol }
      eval "#{symbol}"
    else
      nil
    end
  end

  def initialize(parameters)
    ALL_FIELDS.each do |symbol|
      if parameters[symbol]
        eval("@#{symbol.id2name} = " + parameters[symbol].to_s)
      end
    end
    COMPUTABLE_FIELDS.each do |symbol|
      eval "@#{symbol.id2name} ||= compute_#{symbol.id2name}"
    end
  end

  def *(number)
    TrecFMeasure.new(hashify(LOCAL_FIELDS) { |symbol| field(symbol) * number }.merge(global_field_hash))
  end

  def +(other)
    raise IncompatibleGlobalDataError unless global_field_hash == other.global_field_hash
    TrecFMeasure.new(hashify(LOCAL_FIELDS) { |symbol| field(symbol) + other.field(symbol) }.merge(global_field_hash))
  end

  def ==(other)
    ALL_FIELDS.inject(true) { |equality, symbol| equality && field(symbol) == other.field(symbol) }
  end

  def coerce(other)
    case other
    when TrecFMeasure: return other, self
    when 0: return zero_measure, self
    else
      super
    end
  end

  def zero_measure
    TrecFMeasure.new(hashify(LOCAL_FIELDS) { 0 }.merge(global_field_hash))
  end

  def global_field_hash
    hashify(QUESTION_GLOBAL_FIELDS) { |symbol| field(symbol) }
  end

  def fields
    hashify(ALL_FIELDS) { |symbol| field(symbol) }
  end

  private
  
  def hashify(fields)
    fields.inject(Hash.new) { |hash, symbol| hash[symbol] = yield(symbol); hash }
  end

  def compute_allowance
    100 * (okay_count + vital_count)
  end

  def compute_recall
    (0.0 + vital_count) / num_vital
  end

  def compute_precision
    case
    when length == 0: 0
    when length < allowance: 1
    else
      1.0 - (0.0 + length - allowance) / length
    end
  end

  def compute_f_measure
    if precision < 1E-15 && recall < 1E-15
      0
    else
      ((beta ** 2 + 1.0) * precision * recall)/(beta ** 2.0 * precision + recall)
    end
  end

end
