#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <unistd.h>
#include <limits.h>
#include <float.h>

#define MAX_DF 65535

#ifndef DBL_EPSILON
#define DBL_EPSILON DBL_MIN
#endif

double pdf_t(double x, double df);
double cdf_t(double x, double df);
double t_o(double target_alpha, double df, double epsilon, double*precision);
double df_o(double target_alpha, double target_t, double epsilon, double*precision);

/* < internal junk from the R language library */
#define M_LN_SQRT_2PI .91893853320467274178
#define DOMAIN_ERROR return -1
#define EDOM -1
int errno;

/* < gamma distribution, etc */
int signofgamma = 0; /* the sign of gamma(x), set in lgamma(.), used by gamma(.) */

static double Gam_pos(double);
static double lGam_neg(double);
static double asform(double); /* ASymptotic FORM */

double lgamma(double x)
{
  signofgamma = 1.0;
  if (x <= 0.0) return (lGam_neg(x));
  if (x >  8.0) return (asform(x));
  return (log(Gam_pos(x)));
}

/* Coefficients  from Cheney and Hart (??--REF---??) */
/* Asymptotic form: (not quite, why are the coeff. slightly changed ? */
#define M 6
static double p1[] =
{
  0.83333333333333101837e-1,
  -.277777777735865004e-2,
  0.793650576493454e-3,
  -.5951896861197e-3,
  0.83645878922e-3,
  -.1633436431e-2,
};
static double asform(double x)
{
  /* Equation 6.1.41 Abramowitz and Stegun -- extended Stirling */
  /* See also ACM algorithm 291 */

  double log();
  double nfac, xsq;
  int i;

  xsq = 1. / (x * x);
  for (nfac = 0, i = M - 1; i >= 0; i--) {
    nfac = nfac * xsq + p1[i];
  }
  return ((x - .5) * log(x) - x + M_LN_SQRT_2PI + nfac / x);
}

static double lGam_neg(double x)
{ /* log |gamma(x)|   for  x <= 0   ### G(-x) = - pi / [ G(x) x  sin(pi x) ] */
  double sinpx;
  /* by Gremio:  double log(), sin(), Gam_pos(); */

  x = -x;
  sinpx = sin(M_PI * x);
  if (sinpx == 0.0)
    DOMAIN_ERROR;
  if (sinpx < 0.0)
    sinpx = -sinpx;
        else
	  signofgamma = -1;
  return (-log(x * gamma(x) * sinpx / M_PI));
}

/* Coefficients for rational approximation  gamma(x),  2 <= x <= 3 : */
#define N 8
static double p2[] =
{
  -.42353689509744089647e5,
  -.20886861789269887364e5,
  -.87627102978521489560e4,
  -.20085274013072791214e4,
  -.43933044406002567613e3,
  -.50108693752970953015e2,
  -.67449507245925289918e1,
  0.0,
};
static double q2[] =
{
  -.42353689509744090010e5,
  -.29803853309256649932e4,
  0.99403074150827709015e4,
  -.15286072737795220248e4,
  -.49902852662143904834e3,
  0.18949823415702801641e3,
  -.23081551524580124562e2,
  0.10000000000000000000e1,
};

/* gamma(x) for x >= 0 :*/
static double Gam_pos(double x)
{
  double n, d, s;
  register int i;

  if (x < 2.0) return (Gam_pos(x + 1.0) / x);
  if (x > 3.0) return ((x - 1.0) * Gam_pos(x - 1.0));
  /*-- rational approximation for  2 <= x <= 3 */
  s = x - 2.;
  for (n = 0, d = 0, i = N - 1; i >= 0; i--) {
    n = n * s + p2[i];
    d = d * s + q2[i];
  }
  return (n / d);
}
/* > */

double beta(double a, double b)
{
        return exp(lgamma(a) + lgamma(b) - lgamma(a + b));
}

double pdf_t(double x, double df)
{
  if (df <= 0.0)
    DOMAIN_ERROR;
  return pow(1.0 + x * x / df, -0.5 * (df + 1.0))
    / (sqrt(df) * beta(0.5, 0.5 * df));
}

/* < pbeta */
extern double dbeta(double, double, double);
extern double pbeta(double, double, double);
extern double pbeta_b(double, double, double, double);
extern double qbeta(double, double, double);
extern double rbeta(double, double);
static double incbcf(), incbd(), pseries();

static double big = 4.503599627370496e15;
static double biginv = 2.22044604925031308085e-16;

double lbeta(double a, double b)
{
  return lgamma(a) + lgamma(b) - lgamma(a + b);
}

double pbeta(double x, double a, double b)
{
  return pbeta_b(x, a, b, lbeta(a,b));
}

/* < pbeta_b */
double pbeta_b(double xx, double aa, double bb, double logbeta)
{
  /* logbeta == log(beta(aa,bb)) = log(beta(bb,aa)) */
  double a, b, t, x, xc, w, y;
  int swap_tail;

  if (aa <= 0 || bb <= 0)
    goto domerr;

  if (xx <= 0 || xx >= 1) {
    if (xx == 0) return 0;
    if (xx == 1) return 1;
  domerr:
    errno = EDOM;
    return 0;
  }
  swap_tail = 0;
  if (bb * xx <= 1 && xx <= 0.95) {
    t = pseries(aa, bb, xx, logbeta);
    goto done;
  }
  w = 1 - xx;

  /* Reverse a and b if x is greater than the mean. */

  if (xx > (aa / (aa + bb))) {
    swap_tail = 1;
    a = bb;
    b = aa;
    xc = xx;
    x = w;
  } else {
    a = aa;
    b = bb;
    xc = w;
    x = xx;
  }

  if (swap_tail && (b * x) <= 1 && x <= 0.95) {
    t = pseries(a, b, x, logbeta);
    goto done;
  }

  /* Take __Continued Fraction__  expansion.
   * Choose the one with better convergence.
   *
   * Both are from  Abramowitz & Stegun, sec. 26.5
   */

  if (x * (a + b - 2) < a - 1)
    w = incbcf(a, b, x);
        else
	  w = incbd(a, b, x) / xc; /* A & S, 26.5.9 */

  /* Multiply w by the factor
             a      b   _             _     _
	     x  (1-x)   | (a+b) / ( a | (a) | (b) ) .   */

  y = a * log(x);
  t = b * log(xc);

  /* Resort to logarithms.  */

  y += t - logbeta; /* = lgamma(a + b) - lgamma(a) - lgamma(b); */
  y += log(w / a);
  t = exp(y);

 done:

  if (swap_tail) {
    if (t <= DBL_EPSILON)
      t = 1 - DBL_EPSILON;
                else
		  t = 1 - t;
  }
  return t;
}
/* > */

/* < incbcf */
/* Continued fraction expansion #1.  Abramowitz & Stegun, 26.5.8  */

static double incbcf(double a, double b, double x)
{
  double xk, pk, pkm1, pkm2, qk, qkm1, qkm2;
  double k1, k2, k3, k4, k5, k6, k7, k8;
  double r, rel_err, ans, rel_tol = 3 * DBL_EPSILON;
  int n;
  k1 = a; k2 = a + b;
  k3 = a; k4 = a + 1;

  k5 = 1;         k6 = b - 1;
  k7 = a + 1;     k8 = a + 2;

  pkm2 = 0; qkm2 = 1;
  pkm1 = 1; qkm1 = 1;
  ans = 1;
  r = 1;
  for(n=0; n < 300; n++) {
    xk = -(x * k1 * k2) / (k3 * k4);
    pk = pkm1 + pkm2 * xk;
    qk = qkm1 + qkm2 * xk;
    pkm2 = pkm1; qkm2 = qkm1;
    pkm1 = pk;   qkm1 = qk;

    xk = (x * k5 * k6) / (k7 * k8);
    pk = pkm1 + pkm2 * xk;
    qk = qkm1 + qkm2 * xk;
    pkm2 = pkm1; qkm2 = qkm1;
    pkm1 = pk;   qkm1 = qk;

    if (qk != 0)
      r = pk / qk;
    if (r != 0) {
      rel_err = fabs((ans - r) / r);
      ans = r;
    } else rel_err = 1;

    if (rel_err < rel_tol) break;

    k1 += 1; k2 += 1;
    k3 += 2; k4 += 2;

    k5 += 1; k6 -= 1;
    k7 += 2; k8 += 2;

    /* re-normalize  numerators and denominators if necessary */
    if ((fabs(qk) + fabs(pk)) > big) {
      pkm2 *= biginv; pkm1 *= biginv;
      qkm2 *= biginv; qkm1 *= biginv;
    }
    if ((fabs(qk) < biginv) || (fabs(pk) < biginv)) {
      pkm2 *= big; pkm1 *= big;
      qkm2 *= big; qkm1 *= big;
    }
  }
  return ans;
}
/* > */

/* < incbd */
/* Continued fraction expansion #2.  Abramowitz & Stegun, 26.5.9  */

static double incbd(double a, double b, double x)
{
  double xk, pk, pkm1, pkm2, qk, qkm1, qkm2;
  double k1, k2, k3, k4, k5, k6, k7, k8;
  double r, rel_err, ans, z, rel_tol = 3 * DBL_EPSILON;
  int n;

  k1 = a;         k2 = b - 1;
  k3 = a;         k4 = a + 1;

  k5 = 1;         k6 = a + b;
  k7 = a + 1;     k8 = a + 2;

  pkm2 = 0; qkm2 = 1;
  pkm1 = 1; qkm1 = 1;
  z = x / (1 - x);
  ans = 1;
  r = 1;
  for(n=0; n < 300; n++) {
    xk = -(z * k1 * k2) / (k3 * k4);
    pk = pkm1 + pkm2 * xk;
    qk = qkm1 + qkm2 * xk;
    pkm2 = pkm1; qkm2 = qkm1;
    pkm1 = pk;   qkm1 = qk;

    xk = (z * k5 * k6) / (k7 * k8);
    pk = pkm1 + pkm2 * xk;
    qk = qkm1 + qkm2 * xk;
    pkm2 = pkm1;
    pkm1 = pk;
    qkm2 = qkm1;
    qkm1 = qk;

    if (qk != 0)
      r = pk / qk;
    if (r != 0) {
      rel_err = fabs((ans - r) / r);
      ans = r;
      if (rel_err < rel_tol) break;
    }
    k1 += 1; k2 -= 1;
    k3 += 2; k4 += 2;

    k5 += 1; k6 += 1;
    k7 += 2; k8 += 2;

    /* re-normalize  numerators and denominators if necessary */
    if ((fabs(qk) + fabs(pk)) > big) {
      pkm2 *= biginv; pkm1 *= biginv;
      qkm2 *= biginv; qkm1 *= biginv;
    }
    if ((fabs(qk) < biginv) || (fabs(pk) < biginv)) {
      pkm2 *= big; pkm1 *= big;
      qkm2 *= big; qkm1 *= big;
    }
  }
  return ans;
}
/* > */

/* < pseries */
/* Power series for incomplete beta integral.
 *  Use when b*x is small and x not too close to 1, here  (b*x) <= 1 & x <= 0.95
 */
static double pseries(double a, double b, double x, double logbeta)
{
  /* logbeta == log(beta(a,b)) = log(beta(b,a)) */
  double s, t, u, v, n, t1, z, ai;

  ai = 1 / a;
  u = (1 - b) * x;
  v = u / (a + 1);
  t1 = v;
  t = u;
  n = 2;
  s = 0;
  z = DBL_EPSILON * ai;
  while (fabs(v) > z) {
    u = (n - b) * x / n;
    t *= u;
    v = t / (a + n);
    s += v;
    n += 1;
  }
  s += t1;
  s += ai;

  u = a * log(x);
  t = u + log(s) - logbeta;
  s = exp(t);
  return s;
}
/* > */

/* > */

double cdf_t(double x, double df)
{
  double val = 0.5 * pbeta(df / (df + x * x), df / 2.0, 0.5);
  return (x > 0.0) ? 1 - val : val;
}

/* > */

/* < t_o */
double t_o(double target_alpha, double df, double epsilon, double*precision){
  double lo=-50, hi=50, prev=DBL_MAX;
  double guess = 0, value=0;

  int debug=0;

  if ((target_alpha > 1 - epsilon) ||
      (target_alpha <     epsilon))
    DOMAIN_ERROR;

  while (fabs(prev - guess) > epsilon) {
    prev=guess;
    value=cdf_t(guess, df);
    /*    if (debug) printf("hi=%1.70G\n lo=%1.70G\n guess=%1.70G\n value=%1.70G\n", hi, lo, guess, value);*/
    if (debug) printf("%1.70G\n   %1.70G\n", hi-lo, 0.5*(hi-lo));
    if (target_alpha < value)
      hi=guess;
    else
      lo=guess;
    guess=0.5*(hi+lo); 
  }
  *precision=-log10(fabs(hi-lo));
  return guess;
}
/* > */
/* < df */
double df_o(double target_alpha, double target_t, double epsilon, double*precision){
  double lo=1, hi=MAX_DF, prev=INT_MAX;
  double guess = MAX_DF / 2, value=0;

  int debug=0;

  if ((target_alpha > 1 - epsilon) ||
      (target_alpha <     epsilon))
    DOMAIN_ERROR;
  
  while (fabs(prev-guess) > epsilon) {
    prev=guess;
    value=cdf_t(target_t, guess);
    if (debug) printf("hi=%1.70G\n lo=%1.70G\n guess=%1.70G\n value=%1.70G\n", hi, lo, guess, value);
    if (target_alpha < value)
      hi=guess;
    else
      lo=guess;
    guess=0.5*(hi+lo); 
  }
  *precision=-log10(fabs(hi-lo));
  return guess;
}
/* > */

/* < usage */    
void Usage(char *file){
  printf("usage: %s [-t (-inf,+inf)] [-a (0,1]] [-d (0,+inf)] [-p (0,+inf)] [-h]\n\n", file);
  printf("\t-t\t x value up to which to evaluate t distribution\n");
  printf("\t-a\t alpha value: t_o such that P(t>t_o)=alpha\n");
  printf("\t-d\t degrees of freedom (should be n-1)\n");
  printf("\t-p\t desired precision (actual will be as displayed)\n");
}
/* > */

/*      name           theoretical       actual range       notes          variables
  1     P and Q        (0,1],(0,1]       -----------        SUM to 1.0     p, a
  2     T              (-inf.,inf.)      [-1E300,1E300]                    t
  3     DF             (0,infinity)      [1E-300,1E10]                     df
*/

int main (int argc, char **argv)
{
  /* < which_* flags */
#define which_t  0x1
#define which_a  0x2
#define which_df 0x4
#define which_all 0x7
  /* > */

  char c;
  int	which=0, howmany=0;
  int	errflg=0;
  double a,t,df,prec=0;
  extern int optind;
  extern char *optarg;
  
  /* < command line options */
  if (argc == 1) errflg++;
  while ((c = getopt(argc, argv, "t:d:a:p:h")) != -1) {
    switch (c) {
    case 't':
      t = atof(optarg);
      which |= which_t;
      howmany++;
      break;
    case 'd':
      df = atof(optarg);
      which |= which_df;
      howmany++;
      break;
    case 'a':
      a = atof(optarg);
      which |= which_a;
      howmany++;
      break;
    case 'p':
      prec=pow(10,-atof(optarg)-1);
      break;
    case 'h': errflg++;
      break;
    case '?': errflg++;
    }
  }

  errflg+=abs(howmany-2); /*must have exactly two parameters.*/
  if ((argc - optind) != 0) errflg++;
  if (errflg) { 
    Usage(argv[0]);
    exit(1);
  }
  /* > */

  if (!prec) prec=DBL_MIN;
  /* < range check df */
  if (df > MAX_DF) df=65535;
  if (df <= 0)
    {fprintf(stderr, "degrees-of-freedom overflow\n"); return EDOM;}
  /* > */

  if (!(which & which_t)) {
    t=t_o(a, df, prec, &prec);
    /* < set precision and output field spec */
    howmany=(int)floor(prec+.5);
    if ((howmany > 999) || (howmany < 0)) {fprintf(stderr, "precision overflow\n"); return -1;}
    optarg=calloc(11,sizeof(char));
    if (optarg == NULL) return -1;
    sprintf(optarg, "%%1.%dG\n", howmany);
    /* > */
    printf(optarg, a=t);
  } else if (!(which & which_a)) {
    /* < set precision and output field spec */
    howmany=(int)floor(-log10(prec)+.5);
    if ((howmany > 999) || (howmany < 0)) {fprintf(stderr, "precision overflow\n"); return -1;}
    optarg=calloc(11,sizeof(char));
    if (optarg == NULL) return -1;
    sprintf(optarg, "%%1.%dG\n", howmany);
    /* > */
    printf(optarg, a=cdf_t(t, df));
  }
  else if (!(which & which_df)) {
    df=df_o(a, t, prec, &prec);
    /* < set precision and output field spec */
    howmany=(int)floor(prec+.5);
    if ((howmany > 999) || (howmany < 0)) {fprintf(stderr, "precision overflow\n"); return -1;}
    optarg=calloc(11,sizeof(char));
    if (optarg == NULL) return -1;
    sprintf(optarg, "%%1.%dG\n", howmany);
    /* > */
    printf(optarg, a=df);
  }
	   
  return (a == EDOM);
}





