#!/usr/local/bin/perl -w

use Carp qw(carp croak confess);
#die &round(-38, 1);

#< Constants

$Version=q|$ID$|;
use Getopt::Std;

$ALPHA=0.05;    #for 95% confidence -- .05; for 99% -- .01 you get the idea.
$O_PRECISION=4; #output precision (as in "%1.${O_PRECISION}f")
$MYNAME=`basename $0`;
chop($MYNAME);

$USAGE="
usage: 
$MYNAME -h [-v]
$MYNAME [-a n] [-b] [-p n] [-g] [-s] listfile 
                       column1[:label1[:gran1]]  [column2[:label2[:gran2]]]

 * listfile is mandatory even if it is - for stdin
 * columns contain only digits, labels may contain anything but space and :
 * if either column is specified as 0, only the other will be counted.

Options: 

 -a n  specifies n as alpha.  Default is $ALPHA
 -p n  specifies n as the maximum output format precision.  Default is $O_PRECISION
 -b    asks for brief output: #observations, mean, and confidence interval

       Two graph options for a quick very rough view of the data:

 -g    show histograms of the data.  See also the granularity name options.
       For histograms, l (lower case) means logarithmic buckets, _ 
       (underscore) means log values, and L (upper case) means both. 
 -s    show a scatter plot of two-column data  See also granularity options.
       The scatter plot will show . for one, 2-F for 2-16 points, and * for
       more points at any location.  'l' and 'L' mean log axis, '_' is ignored.
 -S    show the scatter plot using a logarithmic scale for the Z axis, so 
       that 'B' will mean that log(z)=11 for that location.

Labels and Granularities: 

   When specifying the columns, you may specify a label for each, and
   a granularity.  The granularity is the number of buckets to display
   in the graphs requested, and an optional 'L' to indicate
   logarithmic division of the buckets.  '5:foo:11' requests that the
   fifth column be called 'foo' and be shown with 11 evenly-spaced
   buckets.  '7:bar:L' asks the seventh column, labelled 'bar', to be
   shown with the default number of logarithmically-spaced buckets.

The Data:

   Data items may be numeric or time expressions.  Non-numeric data
   are mostly ignored, except some special values (e.g. NaN) that 
   generate warnings.  

   Numeric expressions between zero and one must include the leading
   zero: C<0.25>, not just C<.25>.

   Time may be specified as C<1:23>, C<1:23:45>, or C<1:23:45.67>.
   These are intended as durations, so there is no date processing,
   and the values following a colon (:) are assumed to be out of 60.
   Output for time columns is in the smallest unit, not converted back
   to time form, and a column may mix the two formats.

";

$cdft = "";
if(defined $ENV{"NUGGETEER_HOME"}) {
    $cdft = "$ENV{NUGGETEER_HOME}/blib/bin/cdft";
} else {
    $cdft = "cdft";
}
#>

#< process arguments

$opt_a="";
$opt_b="";
$opt_p="";
$opt_g=undef;
$opt_s=undef;
$opt_S=undef;
my $logarithmic_g = undef;
getopts('hbvgsSp:a:') || die "invalid option: \n$USAGE";
if ($opt_v) {$opt_v=$opt_v; print "version: $Version\n";}
if ($opt_h) {$opt_h=$opt_h; print "$USAGE"; exit;}
if ($opt_p !~ /^$/) {$O_PRECISION=int($opt_p);}	#special match on these two
if ($opt_a !~ /^$/) {$ALPHA=$opt_a;} #to permit '0' argument.

($file, $a, $b) = @ARGV;
#< check that there's a $file
unless ($file) {
    die "#Error: no input file. \n$USAGE";
}
$b ||= 0;
#>

my ($g_agranularity, $g_alog);
my ($g_bgranularity, $g_blog);
#< assign labels and graph granularities
if ($a =~ ":") {
    my $agran;
    ($a, $aname, $agran) = split(":", $a);
    ($g_agranularity) = ($agran =~ /(\d+)/) if $agran;
    ($g_alog) = ($agran =~ /([lL_])/) if $agran;
}
else {
    $aname = "col1";
}
if (($b) && ($b =~ /:/)) {
    my $bgran;
    ($b, $bname, $bgran) = split(":", $b);
    ($g_bgranularity) = ($bgran =~ /(\d+)/) if $bgran;
    ($g_blog) = ($bgran =~ /([lL_])/) if $bgran;
}
else {
    $bname = "col2";
}
$g_alog ||= "";
$g_blog ||= "";
if ($opt_S) {
    $opt_s = 1;
    $g_slog= 1;
}
$g_slog ||= "";
#>

#< is list $a 0?  die "no column spec"
unless ($a) {
    if ($b) {
	$a = $b;
	$aname = $bname;
	$b = 0;
	$bname = "";
    }
    else {die "#Error: no column spec.\n$USAGE\n";}
}
#> 

$PREC=10**$O_PRECISION;

#>

open(FILE, "$file") or die "could not open input file $file.\n";

#< read lists, find difs
$asum=0;
$bsum=0;

$a-=1; # 0-based in perl instead of 1-based counting.  
$b-=1; # will convert back soon.

while (<FILE>) {
    chomp;
    $_ =~ s/\#.*//g;
    @line = split;
    next unless scalar @line;

    my $la;
    my $lb;
    
    warn "WARNING: possibly special value [$line[$a]] ignored at line $.\n"
	if (defined $line[$a] and $line[$a] =~ /^(NaN|Inf|-Inf|undef)$/);
    warn "WARNING: possibly special value [$line[$b]] ignored at line $.\n"
	if ($b>=0 and 
	    defined $line[$b] and $line[$b] =~ /^(NaN|Inf|-Inf|undef)$/);
    $la = $line[$a] 
	if defined $line[$a] and $line[$a] =~ /^((\d+:){0,2}|-?)\d+(\.\d+)?$/;
    $lb = $line[$b] if $b >= 0 and
	defined $line[$b] and $line[$b] =~ /^((\d+:){0,2}|-?)\d+(\.\d+)?$/;
    $la = &sec($la) if defined $la;
    $lb = &sec($lb) if defined $lb;
    
    if ($b >= 0) {
	#< warn and next unless both $la and $lb defined
	if (not defined $la or not defined $lb) {
	    if (defined $la) {
		if (defined $line[$b]) {
		    warn "LINE $.: $bname nonnumeric [$line[$b]]:\n[$_]\n";
		} else {
		    warn "LINE $.: $bname empty where $aname present:\n[$_]\n";
		}
	    } elsif (defined $lb) {
		if (defined $line[$a]) {
		    warn "LINE $.: $aname nonnumeric [$line[$a]]:\n[$_]\n";
		} else {
		    warn "LINE $.: $aname empty where $bname present:\n[$_]\n";
		}
	    } else {
		# both are non-numeric; fine.
	    }
	    next;
	}
	#>

	$bsum += $lb;
	push @bs, $lb;
	push @difs, $la - $lb;
	push @prods, $la * $lb;
    }
    next unless defined $la;

    $asum += $la;
    push @as, $la;
}

# convert back from 0-based counting:
$a++;
$b++;

#>

#< write input parameters
if (not scalar @as) {
    if (not scalar @bs) {
	die "no data for $aname with $bname.\n";
    } else {
	die "no data for $aname.\n";
    } 
}


$n=scalar(@as);
$df=$n-1;
unless ($opt_b) {
    print "File: $file\n";
    print "${aname}-column: $a\n";
    print "${bname}-column: $b\n" if $b;
    print "Number-of-Observations: $n\n";
    print "Degrees-of-Freedom: $df\n";
    print "alpha: $ALPHA\n";
}
#>
    
#< calculate stats

$mu_long1 = &mean(@as);
$mu1 = &t($mu_long1);
$d1 = &t(&median(@as));
$stdev_long1 = &std_deviation(@as);
$v_long1 = &biased_variance(@as);
$v1 = &t($v_long1);
$s1 = &t($stdev_long1);
$ci1 = &t( &ci(@as) ); 
$gmean_long1 = &geometric_mean(@as) unless $opt_b;
$gmean1 = &t($gmean_long1);
if ($b) {
    $mu_long2 = &mean(@bs);
    $mu_longD = &mean(@difs);
    $mu2 = &t($mu_long2);
    $muD = &t($mu_longD);
    $d2 = &t(&median(@bs));
    $dD = &t(&median(@difs));
    $stdev_long2 = &std_deviation(@bs);
    $stdev_longD = &std_deviation(@difs);
    $v_long2 = &biased_variance(@bs);
    $v_longD = &biased_variance(@difs);
    $v2 = &t($v_long2);
    $vD = &t($v_longD);
    $s2 = &t($stdev_long2);
    $sD = &t($stdev_longD);
    $ci2 = &t( &ci(@bs) ); 
    $ciD = &t( &ci(@difs) ); 
    $gmean_long2 = &geometric_mean(@bs) unless $opt_b;
    $gmean2 = &t($gmean_long2);
    #$gmeanD = &t( &geometric_mean(@difs) );

    $confidence = &p(1 - $ALPHA);
    ($tval, $pval) = &pvalue(@difs);

    if ( $mu_long1 != 0 ) {
	$improvement = &p((($mu_long1 - $mu_long2) / $mu_long1));}
    else { 
	$improvement = "NaN";}
    $mu_unpaired = &t($mu_long1-$mu_long2);
    if ($n < 2) { $ci_unpaired = "NaN"; } 
    else {
	$ci_unpaired = &t(&civ($n) * sqrt(($v_long1 / $n)+
					  ($v_long2 / $n)));
    }

    $covariance_long = &mean(@prods) - $mu_long1 * $mu_long2;
    $covariance = &t($covariance_long);

    my $sqrtvarsqrs = sqrt(($v_long1**2) * ($v_long2**2));
    if ($sqrtvarsqrs) {
	$correlation_long = $covariance_long / $sqrtvarsqrs;
    } else {
	$correlation_long = undef;
    }
    $correlation = &t($correlation_long);

    $corrtau_b = &t(kendalls_tau(\@as, \@bs));
    $meansqerr = 0; map { $meansqerr += $_**2 } @difs; $meansqerr /= @difs;
    $rootmse = &t(sqrt($meansqerr));
}
#>

#< write stat output
if ($opt_b) { # brief output
    print qq(n= $n\t);
    print qq($aname:\t$mu1 +- $ci1\t);
    if ($b) {
	print "<" if (($pval < $ALPHA) && ($tval < 0));
	print ">" if (($pval < $ALPHA) && ($tval > 0));
	print "=" if ($pval > $ALPHA / 2);
	print qq(\t$bname:\t$mu2 +- $ci2);
    }
    print qq(\n);
} else {
    print "${aname}-sum: $asum\n";
    if ($b) {print "${bname}-sum: $bsum\n";}
    
    print "$aname:    $mu1 +- $ci1  median: $d1  s: $s1  v: $v1  gm: $gmean1\n";
    if ($b) {
	print "$bname:    $mu2 +- $ci2  median: $d2  s: $s2  v: $v2  gm: $gmean2\n";
	print "--unpaired comparison--\n";
	print "$aname - $bname unpaired: $mu_unpaired +- $ci_unpaired\n";
	print "($aname - $bname)/$aname=\t$improvement\n";
	print "--paired comparison--\n";
	print "$aname - $bname: $muD +- $ciD  median: $dD   s: $sD   v: $vD\n";
	print "covariance: $covariance   correlation: $correlation   tau_b: $corrtau_b   root_mse: $rootmse\n";
	#< Hypothesis testing (short)
	print "H_o (null hypothesis) := $aname = $bname\n";
	printf "t_o: %f    p-value=P(t>|t_o|): %1.6G\n", $tval, $pval;
	print "With $confidence% confidence, $aname ";
	print "<" if (($pval < $ALPHA) && ($tval < 0));
	print ">" if (($pval < $ALPHA) && ($tval > 0));
	print "=" if ($pval > $ALPHA / 2);
	print " $bname";
	#>
	
	#< Hypothesis testing (long)
#    print "H_o (null hypothesis) := $aname = $bname\n";
#    printf "       t_o: %f    p-value=P(t>|t_o|): %1.6G\n", $tval, $pval;
#    print "H_a := $aname <> $bname:\n";
#    print "       Reject H_o with $confidence% confidence iff  P(t > |t_o|) < (alpha / 2)\n";
#    $cmprsn="";
#    #< do not reject at ALPHA / 2
#    if ($pval > ($ALPHA / 2)) {
#	print "       DO NOT reject the null hypothesis.\n";
#
#	#< Less test at alpha
#	if ($tval < 0) {
#	    $rejection="       Reject H_o with $confidence% confidence iff P(t < t_o) < alpha\n";
#	    if ($pval <= $ALPHA) {
#		print "H_a := $bname > $aname\n$rejection";
#		$cmprsn=">=";
#	    }
#	    else {
#		print "H_a := $bname > $aname\n$rejection";
#		print "       DO NOT reject the null hypothesis.\n";
#	    }
#	}
#	#>
#	#< Greater test at alpha
#	else {
#	    $rejection="       Reject H_o with $confidence% confidence iff  P(t > t_o) < alpha\n";
#	    if ($pval < $ALPHA) {
#		print "H_a := $bname > $aname\n$rejection";
#		$cmprsn=">=";
#	    }		
#	    else {
#		print "H_a := $bname < $aname\n$rejection";	    
#		print "       DO NOT reject the null hypothesis.\n";
#	    }
#	}
#	#>
#    }
#    #>
#    #< reject at ALPHA / 2
#    else {
#	print "       REJECT the null hypothesis.\n";
#	if ($tval < 0) {$cmprsn=">";}
#	else {$cmprsn="<";}
#    }
#    print "Thus $bname $cmprsn $aname\n";
#    #>
    #>
	print "\n";
    }
}
#>

if ($opt_g) {
    #< histograms
    for my $dataset ([$g_agranularity, $g_alog, $aname, @as], 
		     [$g_bgranularity, $g_blog, $bname, @bs]) {
	#< settings (width, granularity, log)
	my $screen_width = 80 - 20; # num spaces for stuff before *s

	my ($granularity, $log_opt, $dsname, @vals) = (@$dataset);
	next unless @vals; # if I only had @as for example...
	$granularity ||= 24; # screen height
	$granularity -= 1;

	my $logy_opt = 0;
	$logy_opt = ($log_opt =~ /[_L]/);
	$log_opt = ($log_opt =~ /[lL]/);

	if ($log_opt) {
	    my @posvals = positives(@vals);
	    if (scalar @posvals) {
		if (scalar @posvals < scalar @vals) {
		    warn "WARNING: Discarded ".(scalar(@vals) - scalar(@posvals))." non-positive values for logarithmic histogram of $dsname.\n";
		    @vals = @posvals;
		} else {
		    # great!
		}
	    } else {
		die "Cannot show logarithmic graph of non-positive values.\n";
	    }
	}
	#>
	#< bucket size
	my $max = &max(@vals);
	my $min = &min(@vals);
	#warn "DEBUG: bounds of [".join(", ", sort @vals)."] are $min to $max.\n"; 
	if ($log_opt) {
	    $max = logb($max);
	    $min = logb($min);
	}

	my $degree = &round(logb($max-$min,10) - 1);
	$max = &round($max, $degree);
	$min = &round($min, $degree);
	die "$max < $min???" if $max < $min;
	die "granularity < 0" if $granularity < 0;
	my $bucketsize = ($max - $min) / $granularity;

	die "ERROR: bucketsize=$bucketsize = ($max - $min) / $granularity\n" 
	    unless $bucketsize;
	#> 
	#< label format
	my $bucket_degree = &round(logb($bucketsize,10) - 1);
	my $max_degree = &round(logb($max,10))+1;
	my $print_string = $bucket_degree;
	if ($bucket_degree < 0) {
	    $print_string =~ s/-//;
	    $print_string = "%".($max_degree+$print_string+1).".".$print_string."f";
	} else {
	    $print_string = "%".$max_degree."d";
	}
	#>

	my @buckets = 0 x $granularity;
	
	#< assign each item to a bucket
	foreach my $val (@vals) {
	    if ($log_opt and $val <= 0) {
		warn "ignoring non-positive value because one cannot take its log.\n";
		next;
	    }
	    my $vale = ($log_opt)?(log($val)):$val;
	    my $bucket = sprintf "%d", (($vale - $min) / $bucketsize);
	    confess "ERROR b<0: b=$bucket; v=$vale m=$min bs=$bucketsize" if $bucket < 0;
	    $buckets[$bucket]++;
	}
	#>

	#< scale the bucket to fit the screen
	my $bucketmax = &max(@buckets);
	my $bucketmax_degree = &round(logb($bucketmax,10)+0.5) . "d";
	$bucketmax = logb($bucketmax) if $logy_opt;

	my $elementsize = ($bucketmax < $screen_width)? 1 : 
	    &round($bucketmax / $screen_width);
	#>
	#< print the buckets
	print "histogram for $dsname, *=$elementsize\n";
	my $v = $min;
	for my $i (0 .. $granularity) {
	    $buckets[$i]||=0;
	    printf(">=$print_string  %".$bucketmax_degree."  %s\n", 
		   ($log_opt?exp($v):$v), $buckets[$i],
		   ("*" x (($logy_opt?logb($buckets[$i]):$buckets[$i]) / $elementsize)));
	    $v += $bucketsize;
	}
	#>
    }
    #>
}
if ($opt_s) {
    #< scatter plot
    if (@as and @bs) {
       	#< settings (width, granularity, log)
	$g_agranularity ||= 80 - 10; # screen width - space for labels
	$g_bgranularity ||= 24; # screen height
	my $xgran = $g_agranularity - 1;
	my $ygran = $g_bgranularity - 1;

	my $xlog_opt = ($g_alog =~ /[lL]/); $xlog_opt ||= '';
	my $ylog_opt = ($g_blog =~ /[lL]/); $ylog_opt ||= '';
	if ($xlog_opt) {
	    my @posvals = positives(@as);
	    if (scalar @posvals) {
		if (scalar @posvals < scalar @as) {
		    warn "WARNING: Discarded ".(scalar(@as) - scalar(@posvals))." non-positive values for $aname.\n";
		    @as = @posvals;
		} else {
		    # great!
		}
	    } else {
		die "Cannot show logarithmic graph of non-positive values.\n";
	    }
	}
	if ($ylog_opt) {
	    my @posvals = positives(@bs);
	    if (scalar @posvals) {
		if (scalar @posvals < scalar @bs) {
		    warn "WARNING: Discarded ".(scalar(@bs) - scalar(@posvals))." non-positive values for $bname.\n";
		    @bs = @posvals;
		} else {
		    # great!
		}
	    } else {
		die "Cannot show logarithmic graph of non-positive values.\n";
	    }
	}
	#>

	#< x bucket size
	my $xmax = &max(@as);
	my $xmin = &min(@as);
	if ($xlog_opt) {
	    $xmax = logb($xmax);
	    $xmin = logb($xmin);
	}
	my $xdegree = &round(logb($xmax-$xmin,10) - 1);
	$xmax = &round($xmax, $xdegree);
	$xmin = &round($xmin, $xdegree);
	my $xbucketsize = ($xmax - $xmin) / $xgran;
	confess "ERROR: xbucketsize=$xbucketsize = ($xmax - $xmin) / $xgran\n"
	    unless $xbucketsize > 0;
	#>
	#< y bucket size
	my $ymax = &max(@bs);
	my $ymin = &min(@bs);
	if ($ylog_opt) {
	    $ymax = logb($ymax);
	    $ymin = logb($ymin);
	}
	my $ydegree = &round(logb($ymax-$ymin,10) - 1);
	$ymax = &round($ymax, $ydegree);
	$ymin = &round($ymin, $ydegree);
	my $ybucketsize = ($ymax - $ymin) / $ygran;
	confess "ERROR: ybucketsize=$ybucketsize = ($ymax - $ymin) / $ygran\n"
	    unless $ybucketsize > 0;
	#>
	#< space for x bucket labels
	my $xbucket_degree = &round(logb($xbucketsize,10) - 1);
	my $xmax_degree = &round(logb($xmax,10))+1;
	my $xprint_string = $xbucket_degree;
	if ($xbucket_degree < 0) {
	    $xprint_string =~ s/-//;
	    $xprint_string = "%".($xmax_degree+$xprint_string+1).".".$xprint_string."f";
	} else {
	    $xprint_string = "%".$xmax_degree."d";
	}
	#> 
	#< space for y bucket labels
	my $ybucket_degree = &round(logb($ybucketsize,10) - 1);
	my $max_degree = &round(logb($ymax,10))+1;
	my $print_string = $ybucket_degree;
	if ($ybucket_degree < 0) {
	    $print_string =~ s/-//;
	    $print_string = "%".($max_degree+$print_string+1).".".$print_string."f";
	} else {
	    $print_string = "%".$max_degree."d";
	}
	#> 

	#< initialize buckets
	my @buckets = ();
	for my $j (0 .. $ygran) {
	    $buckets[$j]= [(0) x ($xgran)];
	}
	#>

	#< assign each item to a bucket
	for my $row (0 .. $#as) {
	    my $xbucket = sprintf "%d", 
	    ((($xlog_opt?logb($as[$row]):$as[$row]) - $xmin) / $xbucketsize);
	    confess "ERROR xb<0: xb=$xbucket; v=$as[$row] m=$xmin bs=$xbucketsize" if $xbucket < 0;
	    my $ybucket = sprintf "%d", 
	    ((($ylog_opt?logb($bs[$row]):$bs[$row]) - $ymin) / $ybucketsize);
	    confess "ERROR yb<0: yb=$ybucket; v=$bs[$row] m=$ymin bs=$ybucketsize" if $ybucket < 0;
#	    print "value ($as[$row],$bs[$row]) => -$xmin/$xbucketsize, -$ymin/$ybucketsize ($xbucket,$ybucket)\n";
	    $buckets[$ybucket][$xbucket]++;
	}
	#>

	#< print the buckets
	my %translate = (0=>' ', 1=>'.', 2=>2, 3=>3, 4=>4, 5=>5, 6=>6, 7=>7,
			 8=>8, 9=>9, 10=>'A', 11=>'B', 12=>'C', 13=>'D',
			 14=>'E', 15=>'F', ''=>'?');
	print "\n";
	print "$bname\n";
	my $v = $ymax;
	for my $iinv (0 .. $ygran) {
	    my $i = $ygran - $iinv;
	    #print "print_string=[$print_string]; ydegree=[$ydegree]; ylog_opt=[$ylog_opt];v=[$v]\n";
	    printf(">=$print_string |", 
		   ($ylog_opt?exp($v):$v));
	    foreach my $j (0 .. $xgran - 1) {
		my $val = ($g_slog?&round(logb($buckets[$i][$j])):$buckets[$i][$j]);
		my $mark = $translate{$val};
		$mark ||= "*";
		print $mark;
	    }
	    print "\n";
	    $v -= $ybucketsize;
	}
	print " " x (6+$max_degree);
	print "-" x $xgran;
	print "\n";
	my $left_width = $max_degree-$xmax_degree;
	my $mid_width = $xgran - $xmax_degree - 6;
	printf("   %${left_width}s$xprint_string %${mid_width}s$xprint_string\n", 
	       "", ($xlog_opt?exp($xmin):$xmin), 
	       "", ($xlog_opt?exp($xmax):$xmax));
	print " " x 20;
	print $aname."\n";
	#>
    }
    #>
}


### subroutines ###
#< stat functions

#< &pvalue(@difs)
sub pvalue 
#returns (t_o, p)
#calculates p-value against null hyp. $n, $mu represent standard normal.
#Since @difs should be distributed with standard normal if the two columns
# represent identical distributions, then checking this gives a good paired
# p-value.
{
    my (@list) = @_;
    my ($n, $mu, $s) = (scalar @list, &mean(@list), &std_deviation(@list));
    if (not $s or not $n) {return (0, 0.5); }
    my ($t) = $mu / ($s / sqrt($n));
    $n -= 1;
    my ($at) = abs($t);
    my ($p) = `$cdft -d $n -t $at`;
    chop($p);
    return ($t, .5 - abs($p - .5));
}
#>
#< &civ(@list)
sub civ 
#returns coefficient for 1-$APLHA level of confidence -- e.g. 1.96
{   
    my ($n) = @_;
    return undef if $n <= 1;
    my ($a) = $ALPHA / 2;
    $n -= 1;
    abs(`$cdft -d $n -a $a`);
}
#>
#< &ci(@list)
sub ci
#returns confidence interval at &civ level of confidence
{   my (@list) = @_;
    my ($s) = &std_deviation(@list);
    return undef unless defined $s;
    &civ(scalar @list) * sqrt(($s * $s) / (scalar @list));
}
#>
#< &sum(@list)
sub sum
#returns the sum of the list
{
    my (@vals) = @_;
    my ($sum) = 0;

    foreach(@vals) {
        $sum += $_;
    }
    $sum;
}
#>
#< &prod(@list)
sub prod
#returns the sum of the list
{
    my (@vals) = @_;
    my ($prod) = 1;

    foreach(@vals) {
        $prod *= $_;
    }
    $prod;
}
#>
#< &mean(@list)
sub mean
#returns the mean of the list
{
    my(@vals) = @_;
    &sum(@vals) / (scalar @vals);
}
#>
#< &median(@list)
sub median
#returns median of the list
{
    my(@vals) = @_;

    @vals = sort { $a <=> $b } @vals;   #sort values numerically.
    $vals[@vals / 2];                   #return the middle value.
}
#>
#< &max(@list)
sub max
#returns maximum element of the list
{
    my(@vals) = @_;

    my $max = shift @vals;
    for my $val (@vals) {
	$max = $val if defined $val and $val > $max;
    }
    return $max;
}
#>
#< &min(@list)
sub min
#returns minimum element of the list
{
    my(@vals) = @_;

    my $min = shift @vals;
    for my $val (@vals) {
	$min = $val if defined $val and $val < $min;
    }
    return $min;
}
#>
#< &positives(@list)
sub positives
#returns the sublist of the input that is positive.
{
    return grep { $_ > 0 } @_;
}
#>
#< &std_deviation(@list)

sub std_deviation
#returns standard deviation of the list
{
    my(@vals) = @_;
    return undef if (1 >= scalar @vals);

    my($mean) = &mean(@vals);

    my($sum2) = 0;
    foreach(@vals) {
        $sum2 += ( $_ - $mean ) ** 2;
    }

    sqrt( $sum2 / (scalar(@vals) - 1));
}

#>
#< &biased_variance(@list)
sub biased_variance
# returns the biased variance of the list, which is useful when
# calculating things like covariance and correlation.
{
    my(@vals) = @_;

    my($mean) = &mean(@vals);

    my($sum2) = 0;
    foreach(@vals) {
        $sum2 += ( $_ - $mean ) ** 2;
    }

    sqrt( $sum2 / scalar(@vals) );
}
#>
#< &geometric_mean(@list)
sub geometric_mean {
    my(@vals) = @_;
    my $orig=scalar @vals;
    @vals = positives(@vals);
    return undef if ($orig != scalar(@vals)); # gm no longer makes sense.
    return undef unless @vals;
#    warn "prod(".join(", ", @vals).")=".&prod(@vals)."  ^  1/n=".(1/scalar(@vals)).$/;
    &prod(@vals) ** (1 / scalar @vals);
}    
#>
#< (tau_b, tau, gamma) = &kendalls_tau(\@list, \@list)

# that's actually a bastard of a name.  Kendall's tau is great, but in
# its usual form is not terribly robust to ties.  Theres tau_b which
# is, but for that I have to compute ties at each rank, which sounded
# like a drag, especially as another statistic called gamma (Goodman
# and Kruskal, 1963) is so often cited immediately after explaining
# tau_b which they say is robust to ties, and is almost trivial to
# compute.  My guess is that most people report tau, not caring about
# the few ties that happen.  Well, I'll report natural tau and beta,
# most people will take only one, and things will be fine. 

# From http://www.statsoft.com/textbook/stnonpar.html : 

# The Gamma statistic (Siegel & Castellan, 1988) is preferable to
# Spearman R or Kendall tau when the data contain many tied
# observations. In terms of the underlying assumptions, Gamma is
# equivalent to Spearman R or Kendall tau; in terms of its
# interpretation and computation it is more similar to Kendall tau
# than Spearman R. In short, Gamma is also a probability;
# specifically, it is computed as the difference between the
# probability that the rank ordering of the two variables agree minus
# the probability that they disagree, divided by 1 minus the
# probability of ties. Thus, Gamma is basically equivalent to Kendall
# tau, except that ties are explicitly taken into
# account.


sub kendalls_tau {
    my ($xs, $ys) = @_;
    confess("unexpected arguments: ".Dumper($xs, $ys))
	unless $xs and $ys and "ARRAY" eq ref $xs and "ARRAY" eq ref $ys 
	and @$xs and @$ys;
    confess("kendall's tau requested on non-square table\n")
	unless @$xs == @$ys;
    
    #< get just the ranks by sorting the values.  @sortx and @sorty
    #my %xranks = (); my %yranks = ();
    #my $r=0; foreach my $x (@$xs) { $xranks{$r++} = $x; }
    #   $r=0; foreach my $y (@$ys) { $yranks{$r++} = $y; }

    #my @sortx = sort { $xranks->{$a} <=> $xranks->{$b} } keys %xranks;
    #my @sorty = sort { $yranks->{$a} <=> $yranks->{$b} } keys %yranks;
    #>

    # stolen from Christof Monz, but errors are mine. 
    my $conc = 0; my $disc = 0; 
    my $extrax = 0; my $extray = 0;
    foreach my $j (0 .. $#{$xs}) {
	foreach my $i (0 .. $j-1) {
	    if (  (($xs->[$i] > $xs->[$j]) and ($ys->[$i] > $ys->[$j])) or
		  (($xs->[$i] < $xs->[$j]) and ($ys->[$i] < $ys->[$j])) ) {
		$conc++;
	    } 
	    elsif (($xs->[$i] == $xs->[$j]) and ($ys->[$i] == $ys->[$j])) {
		# do nothing.  no extras.
	    }
	    elsif ($xs->[$i] == $xs->[$j]) { $extray++; }
	    elsif ($ys->[$i] == $ys->[$j]) { $extrax++; }
	    else { $disc++; }
	}
    }
    return (wantarray ? ("NaN","NaN","NaN") : "NaN") 
	unless ($conc + $disc + $extrax) or ($conc + $disc + $extray);
    
    my $n = @$xs;
    my $T0 = $n * ($n - 1) / 2;

    my $kendalls_S = $conc - $disc;
    my $tau = $kendalls_S / $T0;
    my $gamma = $kendalls_S / ($conc + $disc);

    # copied out of 
    #  http://www.uc.edu/sashtml/proc/zompmeth.htm
    # mistakes in translation to perl are mine. 
    # T1 <- sum(xties * (xties - 1))/2
    # T2 <- sum(yties * (yties - 1))/2
    #my $T1 = 0; foreach my $c (values %xties) { $T1 += $c * ($c - 1) } $T1/=2;
    #my $T2 = 0; foreach my $c (values %yties) { $T2 += $c * ($c - 1) } $T2/=2;
    #my $tau_b = $kendalls_S / sqrt( ($T0 - $T1) * ($T0 - $T2) );

    my $tau_b = $kendalls_S / sqrt(($conc+$disc+$extrax) * 
				   ($conc+$disc+$extray));


    return (wantarray) ? ($tau_b, $tau, $gamma) : $tau_b;
}

#                                         a b c d 
# ($tau_b, $tau, $gamma) = &kendalls_tau([1,1,2,2],    # conc = ac,bc,ad,bd
#                                        [2,2,3,4]);   # disc = none

# die "bad tau: $tau" if (abs($tau - .666667) > .001); # tau underestimates (4/6)
# die "bad tau_b: $tau_b" if (abs($tau_b - .8944272) > .001);
# die "bad gamma: $gamma" if (abs($gamma - 1) > .001); #gamma overestimates (4/4)

# my ($tau_b, $tau, $gamma) = &kendalls_tau([1,2,3,4,5,6,7,8, 9,10],
# 					  [2,3,6,5,1,9,4,7,10, 8]);
# die "bad tau: $tau" if (abs($tau - .5111111111111) > .001);
# die "bad tau_b: $tau_b" if (abs($tau_b - .5111111111111) > .001);
# die "bad gamma: $gamma" if (abs($gamma - .5111111111111) > .001);

# ($tau_b, $tau, $gamma) = &kendalls_tau([1,2,3,4,5, 6,7,8, 9,10],
# 				       [2,3,6,5,1,10,4,7,10, 8]);
# die "bad tau: $tau"     if (abs($tau   - .4888889) > .001); 
# die "bad tau_b: $tau_b" if (abs($tau_b - .4944132) > .001);
# die "bad gamma: $gamma" if (abs($gamma - .5000000) > .001); 
# # in R: 
# # > x <- c(4,10,3,1,9,2,6,7,8,5)
# # > y <- c(5,8,6,2,10,3,9,4,7,1)
# # > cor.test(x,y, method="kendall")

# ($tau_b, $tau, $gamma) = &kendalls_tau([1,2,2,2,3,4,5,5,6,7,8,8,9],
# 				       [1,1,2,2,3,4,5,5,5,6,7,7,7]);

# die "bad tau: $tau"     if (abs($tau   - .8717949) > .001);
# die "bad tau_b: $tau_b" if (abs($tau_b - .9512583) > .001);
# die "bad gamma: $gamma" if (abs($gamma - 1       ) > .001);

#>

#< round(num, degree)

sub round { 
    use POSIX;
    my $n=shift; 
    my $d=shift; 
    croak "cannot round undef\n" unless defined $n;
    if ($d) {
	my $norm = $n * (10**(-$d));
	$rnorm = floor($norm+0.5);
	return $rnorm * (10**($d));
    } else {
	return floor($n+0.5);
    }
}
#print &round(.39936, &round(log(.08871)/log(10))) . $/;
#exit 0;

#>
#< logb(val,base)
sub logb {
    my ($val, $base) = @_;
    $base ||= exp(1);
    if (defined $val and $val > 0) {
	return log($val) / log($base);
    } else {
	#croak "cannot take the log of $val\n";
	return 0; # strictly speaking, this is terrible.  Should be -inf
    }
}
#>

#>

#< formatting functions

#< &sec(min:sec.hundr)
sub sec
#returns number of seconds as sec.hundr, having multiplied by minutes
{   my ($time) = @_;
    my ($min) = 0;
    (defined $time) or $time = 0;
    
    if ($time =~ ":") {
	($min, $time) = split(":", $time);
	$time += $min * 60;
    }
    if ($time =~ ":") {
	($min, $time) = split(":", $time);
	$time += $min * 60;
    }
    $time;
}
#>
#< &t($val)

sub t 
#returns %1.${PREC}f formatted value
{
    my ($score) = @_;
    return "NaN" unless defined $score;
    $score *= $PREC;
    $score = round ($score);
    $score / $PREC;
}

#>
#< &p($val)
sub p
#returns 100 * value (i.e. value as a percent)
{
    my ($score) = @_;
    &t(100 * ($score));
}
#>
#>
