#!/usr/bin/perl -w

#< header

package Nuggeteer;

use strict;
use Carp;
use Data::Dumper;
use Getopt::Long;

BEGIN {
    $ENV{NUGGETEER_HOME} ||= ".";
    if (not -e "$ENV{NUGGETEER_HOME}/nuggeteer.pl" or
	not -d "$ENV{NUGGETEER_HOME}/settings") {
	die "NUGGETEER_HOME=[$ENV{NUGGETEER_HOME}] does not contain nuggeteer.pl or its settings.  Please ensure that Nuggeteer has been built and set the NUGGETEER_HOME environment variable.";
    }
    push @INC, qq($ENV{NUGGETEER_HOME}/blib/lib);

    # don't complain about this variable only being seen once in the
    # file.  It's really okay:
    $dbm::BERKELEY=0; $dbm::BERKELEY=0; 
}

use FileHandle;
use Idf qw(fmeas unique); # and idfs and words and nonstopwords, but these are used by reference and should not be imported.
use dbm;
use porter;

use constant learn => "learn";
use constant judge => "judge";
use constant qresult => "qresult";

#>
#< documentation

use constant VERSION => 0.8;  # this is FAR suboptimal, but subversion has no clean way of doing version strings in source.  The best recommendation I've seen so far online is http://svn.haxx.se/users/archive-2006-01/1143.shtml though I think we could get around the export issue by making this substitution part of the src/nuggeteer.pl --> nuggeteer.pl conversion.  That's a TODO or FIXME or whatever.   Note also that it's probably a good idea to put the Nuggeteer version back into the verbose output.  I know it was there once, but it seems to be gone now.
our $VERSION = VERSION;
use constant USAGE => <<"EOT"
usage: nuggeteer.pl --settings <dataset> submittable_file
EOT
    ;
=pod

=head1 NAME

Nuggeteer - automatic estimated nugget-based evaluation

=head1 SYNOPSYS

nuggeteer.pl --settings O2006 your-Other2006-submittable

=head1 DESCRIPTION

Use the RUN-X judgement files for a "complex" question answering task
to approximately score a new run.

Output format by default is tab-delimited, and column headers are
printed in the first line.  This is intended to make it easy to do
further numerical analysis on a per-question basis.


 --help      This message
 --Version   Print the Nuggeteer version and exit

C<Output>:

 --outeval   give output in trec.eval format instead of tab-delimited.
 --prec=s    a printf format string for printing real output values.  %5.3f

 --runtag    runtag to show in the output.  By default inherits from testfile.

 --quiet     normally dots and other indicators go to stderr to indicate
             a judgement qid read (.), or a response recognized as good (+),
             or as not good (-); or guessed as good (*), or as not good (x).
             With --quiet, don't show these.

 --nowarn    even with --quiet, inconsistencies in the input are shown as
             warnings before something reasonable is done (like the offending
             datum being ignored).  Use this option to suppress these warnings.

 --warn      normally, inconsistencies in judgement files are not shown,
             under the assumption that there's nothing you can do about it
             anyway.  Use this option to show these warnings.

 --brief     normally informative information besides scores is printed
             to stdout, such that each line starts with a # sign.  Many
             processing programs ignore these, including genstats, and
             it is easy to remove using 'grep -v ^\#', but this option
             will also suppress output of this additional information.
             Note that this may make it harder to keep track of the
             provenance of saved output files, so this option is advised
             only for command-line convenience.

C<Parameters>:

 --settings  invoke the default settings for your dataset.

             Nuggeteer is distributed with recommended defaults for
             each dataset that it has been trained on.  The settings
             will include values for every other option below,
             including where the data files are to be found and where
             Nuggeteer may store its database for that dataset.  The
             default settings assume that the data files are stored in
             locations relative the C<NUGGETEER_DATA_DIR> environment
             variable, as described in the README.

             During training, we empirically optimized the settings,
             as summarized here:

                       thrshd n b stem weight stop      tau   rmse
 (definition) 2003   -  0.11  2 3  no  count  remove   0.891  0.027
      (other) 2004   -  0.12  2 3  no   idf   keep     0.881  0.037
      (other) 2005   -  0.15  2 3  no   idf   keep     0.919  0.015
      (other) 2006   -  0.14  2 3 yes  count  keep     0.910  0.018
 (reltionshp) R2005  -  0.10  2 3  no  count  remove   1.000  0.010
            ciqa2006 -  0.15  1 3  no   idf   keep     0.901  0.020

             These come from cross-validation experiments whose results
             are detailed in the settings/*.taus files.

             If you want to run Nuggeteer from anywhere other than its 
             distribution directory, please set the C<NUGGETEER_HOME>
             environment variable so that Nuggeteer knows where to find 
             its settings for each dataset.

             If you want to change the settings you use, we recommend
             creating a new settings file naming the variant, and ensuring
             that you specify a different jdb there.  Nuggeteer will complain
             if a jdb was built with different settings than what it is
             currently being run with.

             NOTE: The --settings option silently overrides any of the
             "expert" settings options that precede it on the command line.


=head2 "Expert" options

These options are for those who do not use one of the predefined --settings.

--jdb dbfile
            Save the n-gram classifiers after learning them from the
            judgement files, or use them if the file already exists.

--maxrank n
            Score as if only the top n responses were in the response file.

--maxrid n
            Score as if only the top n responses were in the training files.

--beta=f    beta to use in f-measure between precision and recall.

--threshold t
            In general this specifies a decision method to use when
            approximating judgements for each response-nugget pair.
            You may specify more than one method, but if you do, you must
            also specify a --output file stem, and each method will
            generate output on that stem plus an identifying extension.
            If only one method is specified and --output is not set, then
            scores will be printed on stdout (see also --output, --brief)

            You may specify multiple threshold methods by separating them
            with commas, or by issuing multiple --threshold arguments.

    --threshold [0.0 .. 1.0]
            if t is a number between 0 and 1 inclusive, the decision method
            will use t as a simple global threshold.  The character "t" may
            precede the number, (i.e. --threshold t0.15), and the "t" is 
            always part of the decision method name, as used with --output 
            for example.
    --threshold maxmarg
            use the human-judged response files to train a simple separator
    --threshold art
            use Alexey Radul's Bayesian threshold finder
    --threshold arj
            use Alexey Radul's Bayesian judge
    --external-thresholder "command"
            use an external program for computing thresholds
            api: will pass: --quiet
                 each line of stdin:
                 qid \t nid \t p,o,s,i,t,i,v,e,s \t n,e,g,a,t,i,v,e,s \n
                 expected stdout:
                 {GLOBAL|QUESTION|NUGGET} \t qid \t nid \t threshold \n
    --external-scorer "command"
            use an external program for score finding
            api: will pass: --quiet --answer_keys a,b,c... training_file
                 training_file will be in the same format as stdin for
                 --external_thresholder
                 each line of stdin:
                 either:
                   JUDGE_REQUEST \t qid \t nid1=>score1,nid2=>score2,... \t string
                   expecting stdout:
                   NUGGET_PROBS_LIST \t nid1=>prob1, nid2=>score2, ...
                 or:
                   QRESULT_REQUEST \t qid
                   expecting stdout:
                   QRESULT_RESPONSE \t <tab-separated f-measure numbers>

--output filestem
            Place output files in "filestem.decision-method-name", most useful when
            running with multiple decision methods, so you know which output went 
            where.

--descriptions-only  vs.
--judgements-too
            Learn n-gram classifiers only from the nugget descriptions,
            or also from the judgements assigned to a particular nugget.

--ngrams n
            Build n-gram classifiers using n-grams of length 1 .. n

--idf-weighting  vs.
--count-weighting
            The score that each classifier uses is a weighted n-gram recall
            of the n-grams learned during training.  These parameters choose
            the term weighting strategy.  Count weighting assigns equal weight
            to all terms, whereas idf weighting assigns weight proportional
            to the inverse document frequency of each term.  For a bigram or
            trigram, the weights are summed.

--remove-stopwords  vs.
--keep-stopwords
            When learning n-grams and evaluating responses, should we keep
            or remove stopwords.

--stem  vs.
--no-stem
            When learning n-grams and evaluating responses, should we first
            use the Porter stemmer on each word?


C<Self-Test>:

--exclude!  excludes the testfile from among those to read, if present.
            Use this to make leave-one-out grading easier.

--ensure-known
            If you don't --exclude, but you're testing a file in your
            training set, then all responses should be known.  This
            just makes sure that's the case, dying otherwise, and
            suppresses default warnings about doing this.


=head1 NOTES

=head1 SEE ALSO

=head1 BUGS

=cut
    ;

#>
#< defaults and user interaction

our %TERM_WEIGHTERS = ( "idf" => \&Idf::idfs,
			"count" => sub { return scalar(@_); } );
our %WORD_SELECTORS = ( "keep" => \&Idf::words,
			"remove" => \&Idf::nonstopwords );

unless (caller()) {
    #< components are built according to the options and settings, and

    #  together specify the behavior of elements of the Nuggeteer
    #  scoring system.
    my %cmpt = ();

    # what decision mechanisms will we use to make approximate judgements:
    $cmpt{decision_methods} = {}; # { dmethod_name => coderef }
    $cmpt{output_handles} = {};   # { dmethod_name => filehandle }

    #>
    #< options affect interaction whereas settings affect calculation
    my %opt = ();

    # information learned from the judgement files is saved here so it
    # doesn't have to be recalculated.
    $opt{judgement_db} = "/tmp/nuggeteer-$$.db";

    # specifies the formatting of output.  "tab" means tab-delimited data
    # whereas "trec.eval" means the sort of thing they distribute from NIST.
    $opt{output_format} = "tab";

    # how much precision should the output values show?
    $opt{number_format} = "\%11.9f";

    # what do qids look like?
    $opt{qidre} = '[a-zA-Z\d\.]+';

    # when testing on one of the judgment files, exclude that file and its
    # siblings (those with the same basename) from training.
    $opt{exclude_testfile} = 0;

    # when testing on a judgement file and not excluding that judgement file,
    # ensure that all responses are covered by the judgement set.
    $opt{ensure_known} = 0;

    # maximum number of input responses to score per question.
    # zero indicates no maximum.
    $opt{max_input_rank} = 0;

    # contract with the caller that stderr will be used only for warnings
    # and errors, not for status messages.
    $opt{quiet} = 0;

    # verbose warnings about input data consistency
    $opt{warnings} = 1;

    # contract with caller to only print per-question scores to stdout
    $opt{brief} = 0;

    # Do I produce "debugging" output or not?
    $opt{debug} = $ENV{DEBUG};

    # output filename or filename stem (if multiple scoring methods requested)
    $opt{outputfile} = "-";

    # runtag to give in the output.  If unset, will inherit test file's runtag.
    $opt{runtag} = "";
    #>
    #< settings affect calculation; if all settings are same between
    #  current run and a previously created db, then we can use that db.
    my %settings = ();
    $settings{jfiles} = [];
    $settings{answerkeys} = [];
    $settings{beta} = 3;
    $settings{ngrams} = 0;
    $settings{lm_from_judgements} = 0;
    $settings{stem} = 0;
    $settings{term_weighting} = "idf"; # vs. "count"
    $settings{stopwords} = "keep";
    $settings{max_judgement_rank} = 0;

    #>

    my @threshold_args=();
    my $external_thresholder;
    my $external_scorer;
    my $specified_settings;

    #< GetOptions
    GetOptions ("help" => sub {exec qq(perldoc $0);},
		"Version" => sub {print VERSION;exit 0;},

		"quiet!" => \$opt{quiet},
		"warn+" => \$opt{warnings},
		"nowarn" => sub { $opt{warnings} = 0; },
		"brief!" => \$opt{brief},
		"debug!" => \$opt{debug},
		"outeval" => sub { $opt{output_format} = "trec.eval"; },
		"prec=s" => \$opt{number_format},

		"settings=s" => sub{ 
		    shift;
		    my $name = shift;
		    my $filename = qq($ENV{NUGGETEER_HOME}/settings/$name.pl);
		    die "did not find setting file $filename"
			unless -e $filename;
		    die "cannot read setting file $filename"
			unless -r $filename;
		    eval {
			require $filename;
		    } or die "trouble importing [$filename]: $@";

		    # we will use the following variables only once:
 		    my $foo = \@nuggeteer_settings::decision_methods;
		    $foo = \%nuggeteer_settings::training_settings;
		    $foo = $nuggeteer_settings::jdb;
		    # and these lines remove the warning about that.

		    push @threshold_args, @nuggeteer_settings::decision_methods;
		    %settings = %nuggeteer_settings::training_settings;
		    $opt{judgement_db} = $nuggeteer_settings::jdb;
		    $specified_settings=$name;
		},

		"answerkey=s@" => \@{$settings{answerkeys}},

		"jdb=s" => \$opt{judgement_db},

		"exclude!" => \$opt{exclude_testfile},
		"ensure-known!" => \$opt{ensure_known},
		"maxrank=i" => \$opt{max_input_rank},
		"maxrid=i" => \$settings{max_judgement_rank},

		"threshold=s@" => \@threshold_args,
		"external-thresholder=s" => \$external_thresholder,
		"external-scorer=s" => \$external_scorer,
		"beta=f" => \$settings{beta},

		"descriptions-only" => sub {$settings{lm_from_judgements}=0;},
		"judgements-too" => sub {$settings{lm_from_judgements} = 1;},
		"ngrams=i" => \$settings{ngrams},
		"nograms" => sub{ $settings{ngrams} = 0; },
		"unigrams" => sub { $settings{ngrams} = 1; },
		"bigrams" => sub { $settings{ngrams} = 2; },
		"trigrams" => sub { $settings{ngrams} = 3; },
		"idf-weighting" => sub { $settings{term_weighting} = "idf"; },
		"count-weighting" => sub { $settings{term_weighting} = "count"; },
		"remove-stopwords" => sub { $settings{stopwords} = "remove"; },
		"keep-stopwords" =>  sub { $settings{stopwords} = "keep"; },
		"stem!" => \$settings{stem},

		"outputfile=s" => \$opt{outputfile},
		"number-format=s" => \$opt{number_format},
		"qidre=s" => \$opt{qidre},
		) or die USAGE;
    #>

    if ($specified_settings and not $opt{exclude_testfile}) {
	$opt{judgement_db} =~ s/$$/$specified_settings/g;
    }

    my $testfile = pop @ARGV;
    unless ($specified_settings) {
	my $answerkey = shift @ARGV;
	my @jfiles = @ARGV;
	@ARGV = ();

	$settings{jfiles} = [ &exclude_test_file(\%opt, $testfile, @jfiles) ];
	unshift @{$settings{answerkeys}}, $answerkey;
    }

    unless ("ARRAY" eq ref $settings{jfiles} and 
	    "ARRAY" eq ref $settings{answerkeys} and
	    scalar(@{$settings{answerkeys}}) and 
	    defined $testfile) {
	die USAGE;
    }

    die "cannot learn language models from judgements unless you specify what length of ngrams to learn.\n" if $settings{lm_from_judgements} and not $settings{ngrams};
    #< quick self-tests
    if ($settings{term_weighting} eq "idf") {
	my $s = $TERM_WEIGHTERS{idf};
	die "idf(the) not less than idf(finish)!\n"
	    unless &{$s}("the") < &{$s}("finish");
    }
    if ($settings{stopwords} eq "remove") {
	my $s = $WORD_SELECTORS{remove};
	my $sent = "the quick brown fox jumps over the lazy dog";
	my @w = &{$s}($sent);
	die "what are we removing? [$sent] ==> [".join(" ", @w)."]\n"
	    unless 6 == scalar(@w);
    }
    #>

    &select_decision_method(\%cmpt, $opt{outputfile},
			    $external_thresholder, $external_scorer,
			    @threshold_args);
    warn "no guessing enabled: unknowns will be counted as wrong.\n"
	if (not scalar(keys %{$cmpt{decision_methods}})
	    and not $opt{ensure_known}) and $opt{warnings};

    &nuggeteer(\%opt, \%settings, \%cmpt, $testfile, @ARGV);
}
#>
#< --exclude some judgement files
sub exclude_test_file {
    my ($opt, $testfile, @jfiles) = @_;

    if ($opt->{exclude_testfile}) {
	my $testre = $testfile;
	$testre =~ s/\.[^\.\/]*$//;
	@jfiles = grep !/^$testre/, @jfiles;
    } else {
        if (not defined $opt->{ensure_known} and grep /^$testfile$/, @jfiles) {
	    warn "Running on a training file without excluding it.\n";
	    warn "This is usually done to get `corrected' results,\n";
	    warn "so I'll also ensure that I'm not guessing on any items.\n";
            $opt->{ensure_known} = 1;
            warn "Specify the --ensure-known flag to avoid this message.\n";
        } else {
            # we're not excluding it and we've either specified whether to
            # ensure-known, or more likely the testfile is not among our
            # training files, so we're good to go.
        }
    }
    return @jfiles;
}
#>


#< nuggeteer

sub nuggeteer {
    my ($opt, $settings, $cmpt, $testfile) = @_;

    #< check early that I can read the testfile
    my $testfh = FileHandle->new("<".$testfile)
	or die "cannot read test file [$testfile]: $!\n";
    #>
    #< print settings to output files

    unless ($opt->{brief}) {
	foreach my $dmname (keys %{$cmpt->{output_handles}}) {
	    my $o = $cmpt->{output_handles}{$dmname};
	    print $o qq(\# nuggeteer = [$0]\n);
	    print $o qq(\# starting at: ).localtime()."\n";
	    print $o qq(\# host = $ENV{HOST}\n) if defined $ENV{HOST};
	    print $o qq(\# user = $ENV{USER}\n) if defined $ENV{USER};

	    print $o qq(\# answerkey = [).
		join(", ", @{$settings->{answerkeys}}).qq(]\n);
	    print $o qq(\# testfile = [$testfile]\n);
	    print $o qq(\# maxrank = $opt->{max_input_rank}\n);

	    print $o qq(\# judgement files = [).
		join(", ",@{$settings->{jfiles}}).qq(]\n);
	    print $o qq(\# lm from responses = $settings->{lm_from_judgements}\n);
	    print $o qq(\# ngrams = $settings->{ngrams}\n);
	    print $o qq(\# stopwords policy = $settings->{stopwords}\n);
	    print $o qq(\# term weighting = $settings->{term_weighting}\n);
	    print $o qq(\# stemming = $settings->{stem}\n);
	    print $o qq(\# database = [$opt->{judgement_db}]\n);

	    print $o qq(\# scoring method = $dmname\n);
	    print $o qq(\# beta = $settings->{beta}\n);
	    print $o qq(\# result output format = [$opt->{output_format}]\n);
	}
    }
    #>

    #< a "jdb", or judgement database, is where I keep language models
    # for scoring a run.  Two jdbs should be equal if they were built
    # using the same $settings.  If I have an existing jdb, I should
    # not need to read the judgement files, answer keys, or anything
    # else in order to score.  Nevertheless, I ask the user to specify
    # everything, so I can double-check.

    if (not -e $opt->{judgement_db}) {
	warn "creating $opt->{judgement_db}\n" unless $opt->{quiet};
	my $jdb = start_jdb($opt->{judgement_db}, $settings, $opt);
	die "failed to start jdb\n" unless $jdb and "HASH" eq ref $jdb;
	my ($nresp, $nassg) = read_judgements($jdb, $opt);
	die "failed while reading judgement files\n"
	    unless $jdb and "HASH" eq ref $jdb;
	calculate_training_scores($jdb);
	finish_jdb($jdb);
	print STDERR "read $nresp responses and $nassg assignments.\n"
	    unless $opt->{quiet};
    } else {
	warn "Reading --jdb [$opt->{judgement_db}]...\n" unless $opt->{quiet};
    }
    my $jdb = open_and_check_jdb($opt->{judgement_db}, $settings, $opt);
    die "failed to open jdb [$opt->{judgement_db}]" unless "HASH" eq ref $jdb;
    #>

    #< let each decision method learn what it needs to from the jdb

    my %kb = (); # a knowledge-base for each decision method
    foreach my $dmname (keys %{$cmpt->{decision_methods}}) {
	my $dm = $cmpt->{decision_methods}{$dmname};
	my $oh = $cmpt->{output_handles}{$dmname};
	$kb{$dmname} = &{$dm}(learn, $jdb, $opt, $oh);
    }

    #>
    #< read the testfile
    my $testinfo = read_testfile($opt, $testfile);
    die "failed to read testfile.\n"
	unless $testinfo and "HASH" eq ref $testinfo;
    #>
    #< judge the responses

    foreach my $qid (@{$testinfo->{qids}}) {
	my $responses = $testinfo->{responses}{$qid};
	my $nuggets = $jdb->{_nuggets}{$qid};
	my %scores = ();
	foreach my $rid (keys %$responses) {
	    my $response = $responses->{$rid};

	    my %explanations = ();
	    my (%score) = score($jdb, $qid, $response, \%explanations);
	    $scores{$rid} = \%score;
	}
	foreach my $dmname (keys %{$cmpt->{decision_methods}}) {
	    # make each judgement:
	    my $dm = $cmpt->{decision_methods}{$dmname};
	    my $oh = $cmpt->{output_handles}{$dmname};


	    my $judgements = &{$dm}(judge, $jdb, $opt, $oh, $kb{$dmname}, $qid,
				    \%scores, $responses);


	    foreach my $rid (sort { $a <=> $b } keys %$responses) {
		foreach my $nid (sort {$a <=> $b} keys %$nuggets) {
		    die "no score for qid=$qid rid=$rid nid=$nid\n" . Dumper($scores{$rid})
			unless defined $scores{$rid}{$nid};
		    print_judgement($oh,
				    $jdb, $opt, $qid, $nid, $rid,
				    $judgements->{$rid}{$nid},
				    $scores{$rid}{$nid},
				    $responses->{$rid});
		}
	    }
	}
    }

    #>
    #< score each question

    if ($opt->{output_format} eq "tab") {
	foreach my $dmname (keys %{$cmpt->{decision_methods}}) {
	    my $oh = $cmpt->{output_handles}{$dmname};
	    print $oh ("#".join("\t", qw(qid runtag----- vital okay length
                                    recall prec f)).$/);
	}
    }
    die "something's very wrong" unless "HASH" eq ref $jdb;
    foreach my $qid (@{$jdb->{_qids}}) {
	foreach my $dmname (keys %{$cmpt->{decision_methods}}) {
	    my $dm = $cmpt->{decision_methods}{$dmname};
	    my $oh = $cmpt->{output_handles}{$dmname};
	    my $qresult = &{$dm}(qresult, $jdb, $opt, $oh, $kb{$dmname}, $qid);
	    print_qresult($qid, $qresult, $oh, $opt);
	}
    }

    #>

    return $jdb, $opt, $cmpt, \%kb;
}

#>


#< the Nuggeteer scoring system


# The judgement database looks like this:
# $jdb = { question-id => { nugget-id => [ nugget-instances ] } }, where
# nugget-instance = { _sum => total-weight, _orig => string, n-gram => weight }
#
# So we're set to compute recall for each nugget instance by finding
# matching n-gram instances, and dividing the sum of their weights by
# the sum of the weights in that nugget instance.
#
# We can then compute a nugget score from nugget instance scores by
# multiplying doubts (rather than confidences), so that nugget weight is
#    1 - \sum_{i \in nugget-instances}{ 1 - recall_i }
#
# A special nugget-id "_inform" gives us:
# $jdb = { question-id => { _inform => { n-gram => { i => 0..1,
#                                            nugget-id => 1 } } } }

#< %scores = score(\%ndb, $qid, $string, MODIFIES [\%explanations])
sub score {
    my ($ndb, $qid, $string, $explanations, $no_tresp) = @_;
    #< prerequisites $ndb, $string, $word_selector
    confess qq(strange database).Dumper($ndb)
        unless "HASH" eq ref $ndb;
    confess "no string?\n" unless defined $string;

    my $word_selector = $WORD_SELECTORS{$ndb->{_settings}{stopwords}};
    die "bad word selector subroutine: ".Dumper($word_selector)
	unless "CODE" eq ref $word_selector;
    #>

    my @words = grep { s/\W//g; $_; } &{$word_selector}($string);
    @words = map { porter($_) } @words if $ndb->{_settings}{stem};

    my $nuggets = $ndb->{$qid};
    my $inform = $nuggets->{_inform};

    my %scores = ();

    # have we already seen this response?
    my $tresp = tresp(@words);
    if (not $no_tresp and
	defined $ndb->{$tresp} and defined $ndb->{$tresp}{$qid}) {
	%scores = %{$ndb->{$tresp}{$qid}};
    }

    if ($ndb->{_settings}{ngrams}) {  # make sure we're learning
	#< sum up indicator weights
	$explanations ||= {};
	foreach my $indicator (unique(indicators($ndb->{_settings}{ngrams},
						 @words))) {
	    if (exists $inform->{$indicator}) {
		foreach my $nid (keys %{$inform->{$indicator}}) {
		    next if $nid eq "i";
		    confess qq(jdb does not have an entry for qid $qid\n)
			unless defined $nuggets;

		    # don't guess if we know something about this resp-nug pair
		    next if exists $scores{$nid} and not ref $scores{$nid};

		    confess qq(malformed jdb entry for qid $qid: \n)
			unless "HASH" eq ref $nuggets;

		    $scores{$nid} ||= {};
		    $explanations->{$nid} ||= {};
		    my $i=0;
		    foreach my $ninstance (@{$nuggets->{$nid}}) {
			if (exists $ninstance->{$indicator}) {
			    $scores{$nid}{$i} += ($ninstance->{$indicator} *
						  $inform->{$indicator}{i});
			    $explanations->{$nid}{$i} .=
				" + $indicator: $ninstance->{$indicator}".
				"*$inform->{$indicator}{i}";
			} else {
			    # do nothing, the current instance wasn't here.
			}
			$i++;
		    }
		}
	    } else {
		# can't tell anything from here.
	    }
	}
	#>

	# now scores = { nid => { instance =>
	#                            sum-of(matched-indicator-weights
	#                                     * their-informativenesses) } }
	# so I can give each instance, and thus each nugget, a score.

	#< normalize each sum
	foreach my $nid (keys %scores) {
	    next unless "HASH" eq ref $scores{$nid}; # don't bother known scores
	    # PRODUCT OF DOUBTS:
#	    my $score = 1;
#	    foreach my $instance (keys %{$scores{$nid}}) {
#	        $explanations->{$nid}{$instance} .= " = $scores{$nid}{$instance}";
#	        $score *= 1 - ($scores{$nid}{$instance} /
#			       $nuggets->{$nid}[$instance]{_sum});
#	        $explanations->{$nid}{$instance} .= "; / $nuggets->{$nid}[$instance]{_sum}";
#	    }
#	    $scores{$nid} = 1 - $score;

	    # MAX:
	    my $score = 0;
	    foreach my $instance (keys %{$scores{$nid}}) {
		my $iscore = ($scores{$nid}{$instance} /
			      $nuggets->{$nid}[$instance]{_sum});
		if ($iscore > $score) {
		    $score = $iscore;
		    $explanations->{$nid}{$instance}
		    .= " = $scores{$nid}{$instance}"
			. "; / $nuggets->{$nid}[$instance]{_sum} = $iscore";
		}
	    }
	    $scores{$nid} = $score;
	}
	#>
    }

    #< ensure a score for every nugget, even if we have no data:
    foreach my $nid (keys %{$ndb->{_nuggets}{$qid}}) {
	$scores{$nid} = 0 unless defined $scores{$nid};
    }
    #>

    return %scores;
}
#>
#< \@nids = judge(\%ndb, $qid, $string)

# sub judge {
#     my ($ndb, $qid, $string) = @_;


#     return undef unless $settings->{ngrams};
#     return undef unless "HASH" eq ref $ndb and exists $ndb->{$qid};
#     confess "no string?\n" unless defined $string;
#     confess "strange ndp for qid=$qid:".Dumper($ndb->{$qid})
# 	unless "HASH" eq ref $ndb->{$qid};

#     my $nuggets = $ndb->{$qid};
#     my %explanations = ();
#     my %scores = score($ndb, $qid, $string, \%explanations);

#     # Done: incorporate external judging into Nuggeteer properly
#     if($external_judge) {
# 	submit_external_judgement_request_with_scores($qid, $string, \%scores);
#     }

#     my @assigned_nids = ();
#     my $rank=0;
#     my $min_threshold = 0;
#     # $scores{0} is the score of the background nugget
#     # (relevant if learning from judgements)
#     $min_threshold = $scores{0} if
# 	defined $scores{0} and $scores{0} > $min_threshold;
#     # $threshold is either a scalar global threshold or a reference to
#     # { qid => { nid => threshold, '_' => question-level threshold },
#     #   '_' => global threshold }
#     $min_threshold = $threshold
# 	if not ref $threshold and $threshold > $min_threshold;
#     foreach my $nid (sort {$scores{$b} <=> $scores{$a}} keys %scores) {
# 	my $nugget_threshold = 0;
# 	my $local_threshold = $min_threshold;
# 	if (ref $threshold) {
# 	    if (defined $threshold->{$qid}) {
# 		if(defined $threshold->{$qid}{$nid}) {
# 		    $nugget_threshold = $threshold->{$qid}{$nid};
# 		} elsif (defined $threshold->{$qid}{_}) {
# 		    $nugget_threshold = $threshold->{$qid}{_};
# 		} else {
# 		    $nugget_threshold = $threshold->{_};
# 		}
# 	    } else {
# 		$nugget_threshold = $threshold->{_};
# 	    }
# 	}
# 	$local_threshold = $nugget_threshold
# 	    if $nugget_threshold > $local_threshold;
# 	$rank++;
# 	# Since the thresholds can change, this optimization is only valid
# 	# against the minimum threshold, not the current local one
# 	last unless $scores{$nid} > $min_threshold # cutoff at threshold
# 	    or ($opt->{debug} and $opt->{debug} > 1);
# 	print qq(\#DEBUG $qid\t$rank\t$nid\t$scores{$nid}\t$string).
# 	    qq(\tNUGGET $nid\t$nuggets->{$nid}[0]{_orig} );
#  	if ($scores{$nid} > $local_threshold) {
#  	    print "ASSIGNED, cutoff was $local_threshold\n";
#  	} else {
# 	    print "NOT ASSIGNED, cutoff was $local_threshold\n";
#  	}
# 	last unless $nid # background matched better than later nuggets
# 	    or $opt->{debug};
# 	push @assigned_nids, $nid if $scores{$nid} > $local_threshold;
# 	if ($opt->{debug} and $opt->{debug} > 2) {
# 	    print qq(\#DEBUG score    $nid = $scores{$nid}\n);
# 	    foreach my $instance ( keys %{$explanations{$nid}} ) {
# 		$explanations{$nid}{$instance} =~ s/(\d+\.\d\d\d\d)\d+/$1/g;
# 		print qq(\#DEBUG x        ).$explanations{$nid}{$instance}.$/;
# 		print qq(\#DEBUG n        $nuggets->{$nid}[$instance]{_orig}\n)
# 	    }
# 	    print qq(\n\n);
# 	}
#     }

#     return \@assigned_nids;
# }

#>
#< learn_string(\%jdb, $qid, $nid, $string)
sub learn_string {
    my ($jdb, $qid, $nid, $string) = @_;
    confess "no jdb!!?\n" unless $jdb and "HASH" eq ref $jdb;
    my $word_selector = $WORD_SELECTORS{$jdb->{_settings}{stopwords}};
    die "bad word selector subroutine: ".Dumper($word_selector)
	unless "CODE" eq ref $word_selector;
    my $term_weighter = $TERM_WEIGHTERS{$jdb->{_settings}{term_weighting}};
    die "bad term weight subroutine: ".Dumper($term_weighter)
	unless "CODE" eq ref $term_weighter;
    my $question = $jdb->{$qid};
    $question ||= {_inform => {}};
    my $nugget;
    if (exists $question->{$nid}) {
	$nugget = $question->{$nid};
    } else {
	$nugget = [];
	$question->{$nid} = $nugget;
    }
    my $ninstance = {_sum => 0, _orig => $string};
    my $inform = $question->{_inform};
    my @words = grep { s/\W//g; $_; } &{$word_selector}($string);
    my @indicators = indicators($jdb->{_settings}{ngrams},@words);

    #< update counts and informativeness
    foreach my $indicator (@indicators) {
	my $weight = &{$term_weighter}(split /\s/, $indicator);
	$indicator = lc($indicator); # we fold case AFTER we get the idf
	$indicator = porter($indicator) if $jdb->{_settings}{stem}; # stem AFTER we get the idf
	next if exists $inform->{$indicator}{-1}; # question word

	$ninstance->{$indicator} = $weight;
	$ninstance->{_sum} += $weight;

	$inform->{$indicator} ||= {i => 0};
	if (exists $inform->{$indicator}{$nid}) {
	    # do nothing: no need to update it or update the informativeness
	} else {
	    $inform->{$indicator}{$nid}=1;
	    $inform->{$indicator}{i} = 1 -
		# number of nuggets this indicator appears in:
		(scalar(keys(%{$inform->{$indicator}})) - 1) / # -1 for i
		# divided by total number of nuggets for this question:
		(scalar(keys(%$question)) - 1);    # -1 for _inform

	    $inform->{$indicator}{i} = 0
		if (exists $inform->{$indicator}{-1}); # question word
	}
    }
    #>

    push @$nugget, $ninstance;
    $jdb->{$qid} = $question;

    my $tresp = "\t" . join(" ", @words);
    return $tresp;
}
#>
#< indicators(@words) -- gather n-and-fewer-gram indicators from @words
sub indicators {
    my ($ngrams, @words) = @_;
    my @indicators = ();
    foreach my $ng (0 .. $ngrams) {
	foreach my $i ($ng..$#words) {
	    push @indicators, join(" ",@words[($i-$ng) .. $i]);
	}
    }
    return @indicators;
}
#>
#< tresp(\%jdb, response, [@words])
sub tresp {
    my ($jdb, $response) = @_;
    if ("HASH" eq ref $jdb) { # otherwise it's being called on a list of words
	my $word_selector = $WORD_SELECTORS{$jdb->{_settings}{stopwords}};
	die "bad word selector subroutine: ".Dumper($word_selector)
	    unless "CODE" eq ref $word_selector;

	my @words = grep { s/\W//g; $_; } &{$word_selector}($response);
	@words = map { porter($_) } @words if $jdb->{_settings}{stem};

	return "\t" . join(" ", @words);
    } else {
	return "\t" . join(" ", @_);
    }
}
#>

#>
#< jdbs

#< @qids = read_answer_key(keyfile, MODIFIES \%questions, MODIFIES \%nuggets, MODIFIES \%vitoks, \%opt)
#  fills the hashes based on the info in the keyfile.
sub read_answer_key {
    my ($nuggetlistfile, $questions, $nuggets, $vitoks, $opt) = @_;
    confess qq(read_answer_key usage: answer_key, \%questions, \%nuggets, \%vitoks, \%opt) unless defined $opt;
    open(NUGGETS, "<", $nuggetlistfile)
	or die "cannot read answer key [$nuggetlistfile]: $!\n";
    print STDERR "reading [$nuggetlistfile]\n"
	if $opt->{debug} and not $opt->{quiet};

    my @qids;

    while (<NUGGETS>) {
	next unless /\S/;
	if (/^Qid ($opt->{qidre}):(\s+'other' question for target)? (.*)/) {
	    $questions->{$1} = $3;
	    push @qids, $1;
	    next;
	}
	chomp;
	my ($qid, $nid, $vitok, $string)
	    = (/^($opt->{qidre})\s(\d+)\s+(vital|okay|[0-9\.]+)\s+(.*)/);
	die "unexpected format [$_]\n" unless defined $string;
	$nuggets->{$qid}||={};
	$vitoks->{$qid} ||={};
	$vitok = 0 if $vitok eq "okay";  # pyramid scoring does the right thing
	$vitok = 1 if $vitok eq "vital"; # with vital/okay if they are 1/0.
	if (exists $nuggets->{$qid}{$nid}) {

	    # NOTE: Multiple descriptions are an extension of the standard
	    #       format, so it's not clear how to handle them.
	    #       Here I choose to store all of them, with the highest
	    #       vitok.  Another reasonable strategy might be to die.
	    #       A third might be to try to form a pyramid, but that is
	    #       likely to be wrong if e.g. one of the files already was
	    #       a pyramid structure.

	    $nuggets->{$qid}{$nid}.="\t".$string;
	    if (exists $vitoks->{$qid}{$nid} and
		$vitok ne $vitoks->{$qid}{$nid}) {
		warn qq(mismatched vitok [$qid $nid] at $nuggetlistfile:$.\n)
		    if $opt->{warnings};
		$vitok = $vitoks->{$qid}{$nid} if
		    $vitoks->{$qid}{$nid} > $vitok;  # choose the greater
	    }
	} else {
	    $nuggets->{$qid}{$nid}=$string;
	}
	$vitoks->{$qid}{$nid} = $vitok;
    }
    print STDERR "read nuggets for ".scalar(keys %$nuggets)." qids from [$nuggetlistfile]\n"
	unless $opt->{quiet};
    return @qids;
}
#>
#< \%jdb = open_and_check_jdb(jdb-filename, \%settings)

sub open_and_check_jdb {
    my ($jdbfile, $settings) = @_;
    my $jdb = readdb($jdbfile);
    if (not defined $jdb->{_closed} or not defined $jdb->{_settings} or
	not defined $jdb->{_questions} or not defined $jdb->{_nuggets} or
	not defined $jdb->{_vitoks} or not defined $jdb->{_qids} or
	not defined $jdb->{_known_pos} or not defined $jdb->{_known_neg}) {
	die "The judgement database [$jdbfile] is unfinished or was created by an incompatible nuggeteer.  If another nuggeteer is running with the same settings, you might need to wait until it is done, just this first time.  Otherwise please move it away and try again.\n";
    }
    if ($jdb->{_settings}{lm_from_judgements}
	eq $settings->{lm_from_judgements} and
	$jdb->{_settings}{stem} eq $settings->{stem} and
	$jdb->{_settings}{ngrams} == $settings->{ngrams} and
	$jdb->{_settings}{stopwords} eq $settings->{stopwords} and
	$jdb->{_settings}{term_weighting} eq $settings->{term_weighting}
	) {

	my $okay=1;
	my @dbjfiles = sort @{$jdb->{_settings}{jfiles}};
	foreach my $jfile (sort @{$settings->{jfiles}}) {
	    my $dbj = shift @dbjfiles;
	    unless ($dbj eq $jfile) { $okay=0; last; }
	}
	die "[$jdbfile] was built from a different set of judgement files: [".join(", ", @dbjfiles)."]\n"
	    unless $okay;

	$okay = 1;
	my @dbanskeys = sort @{$jdb->{_settings}{answerkeys}};
	foreach my $anskey (sort @{$settings->{answerkeys}}) {
	    my $dbak = shift @dbanskeys;
	    unless ($dbak eq $anskey) { $okay=0; last; }
	}
	die "[$jdbfile] was built from a different set of answer keys: [".join(", ", @dbanskeys)."]\n"
	    unless $okay;


    } else {
	die "[$jdbfile] was built with different settings: ".
	    Dumper($jdb->{_settings});
    }
    return $jdb;
}

#>
#< \%jdb = start_jdb(jdbfile, \%settings, \%opt);
sub start_jdb {
    my ($jdbfile, $settings, $opt) = @_;
    confess qq(start_jdb usage: filename, \%settings, \%opt\n)
	unless (defined $jdbfile and not ref $jdbfile and
		defined $settings and "HASH" eq ref $settings and
		defined $opt and "HASH" eq ref $opt);

    my %questions = (); # { qid => "text of each question" }
    my %nuggets = (); #   { qid => { nid => "description1\td2\t...\tdN" } }
    my %vitoks = ();  #   { qid => { nid => [0...1] } }
    my @qids = ();
    foreach my $answerkey (@{$settings->{answerkeys}}) {
	my @q = read_answer_key($answerkey, \%questions, \%nuggets, \%vitoks, $opt);
	@qids = @q unless @qids; # take qids only from the first (main) answerkey.
    }

    my $jdb = writedb($jdbfile);
    $jdb->{_file} = $jdbfile;# if $dbm::BERKELEY;
    die "cannot write judgement database [$jdbfile]: $!$@\n" unless $jdb;
    $jdb->{_questions} = \%questions;
    $jdb->{_nuggets} = \%nuggets;
    $jdb->{_vitoks} = \%vitoks;
    $jdb->{_qids} = \@qids;
    $jdb->{_settings} = $settings;
    if ($settings->{ngrams}) {
	foreach my $qid (keys %nuggets) {
	    learn_string($jdb, $qid, -1, $questions{$qid});
	    foreach my $nid (keys %{$nuggets{$qid}}) {
		foreach my $ntxt (split /\t/, $nuggets{$qid}{$nid}) {
		    print STDERR "~" if $opt->{debug} and not $opt->{quiet};
		    learn_string($jdb, $qid, $nid, $ntxt);
		}
	    }
	}
    }
    return $jdb;
}
#>
#< read_judgements(MODIFIES \%jdb, \%opt);
sub read_judgements {
    my ($jdb, $opt) = @_;
    unless (scalar @{$jdb->{_settings}{jfiles}}) {
	warn "No judgement files.\n" if $opt->{warnings};
	return (0,0);
    }
    confess ("jdb's max judgement rank setting is undefined.")
	unless defined $jdb->{_settings}{max_judgement_rank};

    my $nuggets = $jdb->{_nuggets};  # { qid => { nid => description } }
    $jdb->{_known_pos}||={}; # { qid => { nid => { response => 1 } } }
    $jdb->{_known_neg}||={}; # { qid => { nid => { response => 1 } } }

    # things we might expect in a judgement file:
    use constant question => "Submission for question dddd with items numbered";
    use constant response => "qid runtag rid docid response";
    use constant nuggetsline => "Nuggets assigned to items for question dddd";
    use constant assignment => "qid runtag rid nid";

    my $nresp=0;
    my $nassg=0;
    foreach my $judgement_file (@{$jdb->{_settings}{jfiles}}) {

	my $jfname = $judgement_file;
	$jfname =~ s|.*/||g;
	print STDERR "  learning $jfname" unless $opt->{quiet};
	open(J, "<", $judgement_file)
	    or die "cannot read judgement file [$judgement_file]: $!\n";
	my $expecting = question;
	my $current_qid = 0;
	my %current_responses = (); # { rid => response text }
	my %current_docids = ();    # { rid => response docid }
	my %current_assignments = (); # { nid => { rid => 1 } }
	my $line=0;

	while (<J>) {
	    $line++;
	    chomp;
	    next if /^\#/; # ignore comments
	    #< expecting a question
	    if ($expecting eq question) {
		next if (/^\s*$/);
		if (/^Submission for question ($opt->{qidre}) with items numbered:\s*$/) {
		    $current_qid = $1;
		    %current_responses = ();
		    %current_docids = ();
		    %current_assignments = ();
		} else {
		    die "Bad judgement format at [$judgement_file] line $line: expected [$expecting] but got [$_]\n";
		}
		$expecting = response;
	    }
	    #>
	    #< expecting a response
	    elsif ($expecting eq response) {
		if (/^\s*$/) {
		    $expecting = nuggetsline;
		    next;
		}
		my ($qid, $rid, $docid, $response) =
		    (/($opt->{qidre})\s+\S+\s+(\d+)\s+(\S+)\s+(.*)$/);
		die "Bad judgement format at [$judgement_file] line $line: expected [$expecting] but got [$_]\n"
		    unless defined $docid;
		$current_responses{$rid} = lc(trim($response));
		$current_docids{$rid} = $docid;
		$nresp++;

		if (($jdb->{_settings}{max_judgement_rank} > 0) and
		    ($jdb->{_settings}{max_judgement_rank} <= $rid)) {
		    while (!/^\s*$/) {
			$_ = <J>;
		    }
		    $expecting = nuggetsline;
		    next;
		}
	    }
	    #>
	    #< expecting a nuggetsline
	    elsif ($expecting eq nuggetsline) {
		next if (/^\s*$/);
		if (/^Nuggets assigned to items for question ($opt->{qidre}):\s*$/) {
		    die "At [$judgement_file] line $line: expected nuggets for qid [$current_qid] but found qid [$1] instead!\n"
			unless $1 eq $current_qid;
		} else {
		    die "Bad judgement format at [$judgement_file] line $line: expected [$expecting] but got [$_]\n";
		}
		$expecting = assignment;
	    }
	    #>
	    #< expecting an assignment

	    elsif ($expecting eq assignment) {
		if (/^\s*$/) {
		    add_judgement_set($jdb, $opt,
				      $current_qid, \%current_responses,
				      \%current_docids, \%current_assignments);
		    print STDERR "." unless $opt->{quiet};
		    $expecting = question;
		    next;
		}
		my ($qid, $rid, $nid) =
		    (/($opt->{qidre})\s+\S+\s+(\d+)\s+(\d+)\s*$/);
		die "Bad judgement format at [$judgement_file] line $.: expected [$expecting] but got [$_]\n"
		    unless defined $nid;

		die "unknown qid nid pair [$qid $nid] at $.\n"
		    unless (exists $nuggets->{$qid} and
			    exists $nuggets->{$qid}{$nid});
		if ($jdb->{_settings}{max_judgement_rank} > 0 and
		    $rid > $jdb->{_settings}{max_judgement_rank}) {
		    warn "nugget $nid assigned in $qid for at rank > $jdb->{_settings}{max_judgement_rank}\n" unless $opt->{quiet};
		    next;
		}
		die "no response text for $qid.rid=$rid in ".
		    Dumper(\%current_responses)
		    unless exists $current_responses{$rid};

		$current_assignments{$nid} ||= {};
		$current_assignments{$nid}{$rid} = 1;
		$nassg++;
	    }

	    #>
	}
	add_judgement_set($jdb, $opt, $current_qid, \%current_responses,
			  \%current_docids, \%current_assignments)
	    if $expecting eq assignment; # no blank line at end of file.
	die "Judgement file [$judgement_file] ended prematurely: expected [$expecting], but found [$_]"
	    if $expecting eq response or $expecting eq nuggetsline;
        print STDERR "\n" unless $opt->{quiet};
    }

    return $nresp, $nassg;
}

#>
#< add_judgement_set(MODIFIES \%jdb, qid, responses, docids, assignments)

sub add_judgement_set {
    my ($jdb, $opt, $qid, $responses, $docids, $assignments) = @_;
    my $nuggets = $jdb->{_nuggets}{$qid};  # { nid => description }
    my $lm_from_judgements = $jdb->{_settings}{lm_from_judgements};

#     my %scores = (); # { response => { nid => [0..1] } }
#     foreach my $rid (keys %$responses) {
# 	$scores{$responses->{$rid}} = { score($jdb, $qid, $responses->{$rid}) };
#     }
    
    my $kn = $jdb->{_known_neg};
    my $kp = $jdb->{_known_pos};

    foreach my $nid (keys %$nuggets) {
	if (defined $assignments->{$nid}) {
	    # we have a positive judgement for it.  Nothing else can be
	    # considered negative for this nugget because assessors only
	    # marked the "best" one.
	    foreach my $rid (keys %{$assignments->{$nid}}) {
		my $response = $responses->{$rid};
		#< warn if $kn->{$qid}{$nid}{$response} because it's inconsistent: delete negative
		if (defined $kn->{$qid} and
		    defined $kn->{$qid}{$nid} and
		    defined $kn->{$qid}{$nid}{$response}) {
		    warn "Warning: inconsistent judgement for qid=$qid;rid=$rid;nid=$nid: [$response], removing the negative judgment.\n" if $opt->{warnings} > 1 or $lm_from_judgements;
		    warn "Warning: may already have learned n-grams for the negative judgement!!  This may lead to inconsistent behavior!\n" if $lm_from_judgements;
		    delete $kn->{$qid}{$nid}{$response};
		}
		#>

		learn_string($jdb, $qid, $nid, $response)
		    if $lm_from_judgements;

		#< $kp->{$qid}{$nid}{$response} = 1
		$kp->{$qid}||={};
		$kp->{$qid}{$nid}||={};
		$kp->{$qid}{$nid}{$response}=1;
		#>
		#< $jdb->{tresp($jdb, $response)}{$qid}{$nid} = 1
		my $tresp = tresp($jdb, $response);
		my $kt = $jdb->{$tresp};
		$kt ||= {};
		$kt->{$qid} ||= {};
		$kt->{$qid}{$nid} = 1;
		$jdb->{$tresp} = $kt; # temp assign and reinsert for dbm.
		#>
	    }
	} else {
	    # all responses are negative for this nugget.
	    foreach my $rid (keys %$responses) {
		my $response = $responses->{$rid};
		#< warn if $kp->{$qid}{$nid}{$response} because it's inconsistent: skip this
		if (defined $kp->{$qid} and
		    defined $kp->{$qid}{$nid} and
		    defined $kp->{$qid}{$nid}{$response}) {
		    warn "Warning: inconsistent judgmenent for qid=$qid;rid=$rid;nid=$nid: [$response], ignoring the negative judgment.\n" if $opt->{warnings} > 1;
		    next;
		}
		#>

		learn_string($jdb, $qid, 0, $response)
		    if $lm_from_judgements;

#		if ($scores{$response}{$nid} > 0) { # drop zero-score negatives.
		    #< $kn->{$qid}{$nid}{$response} = 1
		    $kn->{$qid}||={};
		    $kn->{$qid}{$nid}||={};
		    $kn->{$qid}{$nid}{$response}=1;
		    #>
		    #< $jdb->{tresp($jdb, $response)}{$qid}{$nid} = 0
		    my $tresp = tresp($jdb, $response);
		    my $kt = $jdb->{$tresp};
		    $kt ||= {};
		    $kt->{$qid} ||= {};
		    $kt->{$qid}{$nid} = 0;
		    $jdb->{$tresp} = $kt; # temp assign and reinsert for dbm
		    #>
#		}
	    }
	}
    }

    $jdb->{_known_neg} = $kn; # dbms don't like internal assignment, so I
    $jdb->{_known_pos} = $kp; # have to temp assign and reinsert.
}

#>
#< calculate_training_scores(MODIFIES \%jdb)

sub calculate_training_scores {
    my ($jdb) = @_;
    confess qq(calculate_training_scores usage: \%jdb\n)
	unless "HASH" eq ref $jdb;

    # NOTE: I'm not sure how kosher the interaction between this and
    # learn_lm_from_judgements is, because ideally I think I'd like to
    # ablate the current response from the judgement training, and I
    # can't.  I would have to store the provenance of each n-gram in
    # the jdb and test whether there was one and only one response
    # that motivated it, and then whether it was this one.  (I suppose
    # I could only store this info for responses that only had one
    # provenance, as yet, but still).

    my $kp = $jdb->{_known_pos};
    my $kn = $jdb->{_known_neg};
    foreach my $h ($kp, $kn) {
	foreach my $qid (keys %$h) {
	    foreach my $nid (keys %{$h->{$qid}}) {
		foreach my $response (keys %{$h->{$qid}{$nid}}) {
		    my (%score) = score($jdb, $qid, $response, {}, "no tresp");
		    $score{$nid} = 0 unless defined $score{$nid};
		    $h->{$qid}{$nid}{$response}=$score{$nid};
		}
	    }
	}
    }
    $jdb->{_known_pos} = $kp;
    $jdb->{_known_neg} = $kn;
#    warn Dumper $kp, $kn;
}

#>
#< finish_jdb(MODIFIES \%jdb)
sub finish_jdb {
    my ($jdb) = @_;
    $jdb->{_closed} = 1;
    closedb($jdb->{_file}) if exists $jdb->{_file};
    untie $jdb;
}
#>
#< write_training_file(\%jdb, filename)
sub write_training_file {
    my ($jdb, $training_file) = @_;
    open(EXTERNTHRESH, ">", $training_file)
	or die "cannot write temporary file [$training_file]: $!\n";

    foreach my $qid (@{$jdb->{_qids}}) {
	foreach my $nid (sort { $a <=> $b }
			 keys %{$jdb->{_nuggets}{$qid}}) {

	    my @negatives = ();
	    if (defined $jdb->{_known_neg}{$qid}{$nid}) {
		@negatives = reverse sort {$a <=> $b} values
		    %{$jdb->{_known_neg}{$qid}{$nid}};
	    }

	    my @positives = ();
	    if (defined $jdb->{_known_pos}{$qid}{$nid}) {
		@positives = sort {$a <=> $b} values
		    %{$jdb->{_known_pos}{$qid}{$nid}};
	    }

#	    print STDERR qq($qid\t$nid\tpos:\t).join(",", @positives).
#		"\tneg:\t".join(",", @negatives)."\n";
	    print EXTERNTHRESH qq($qid\t$nid\t).join(",", @positives).
		"\t".join(",", @negatives)."\n";
	}
    }
    close EXTERNTHRESH;
}
#>


#>
#< decision methods

# A "decision method" must support three operations:
#
# We promise to call learn first: \%knowledge = dm(learn, \%jdb, \%opt, \*ofh)
#
# Then:  { rid => { nid => 0 or 1 } }
#         = dm(judge, \%jdb, \%opt, \*output_fh, \%knowledge, $qid,
#              \%scores, \%responses)
#        where the %scores are { rid => { nid =>  0<=x<=1  } }
#        and %responses are    { rid => string }
#        for every response-id and nugget-id pair.
#
# Finally: \%qresult = dm(qresult, \%jdb, \%opt, \*out_fh, \%knowledge, $qid);
# A qresult must contain:
#                 vital/okay interpretation         pyramid interpretation
# vital_match - the number of vital matches      sum of weights
# vital_poss  - num possible vital nuggets       sum of possible weights
# okay_match  - the number of okay nuggets       number of nuggets matched
# okay_poss   - num possible okay nuggets
# total_match - vital_match + okay_match         okay_match
# total_poss  - vital_poss + okay_poss           vital_poss
# length      -      sum of non-whitespace characters in all responses
# recall      -                  vital_match / vital_poss
# allowance   -                     100 * total_match
# precision   - if (length < allowance), 1, otherwise: 1 - (len - alwc)/len
# fmeasure    -         F(beta=opt->{beta}) of precision and recall.

#< select_decision_method(MODIFIES \%cmpt, $opt{outputfile}, threshold args)

sub select_decision_method {
    my ($cmpt, $outputfile, $external_thresholder, $external_scorer, @targs)
	= @_;

    # separate commas:
    my @threshold_args = ();
    foreach my $t (@targs) {
	push(@threshold_args, split(/,/, $t));
    }
    push @threshold_args, "1" unless @threshold_args; # exact scoring method

    # $cmpt{decision_methods} = { name => \&coderef }
    # $cmpt{output_handles}   = { name => output_handle }
    # $cmpt{collect_training} = true or false

    if (defined $external_thresholder) {
	my $name = $external_thresholder;
	$name =~ s/[^\w.]+/-/g;
	$cmpt->{decision_methods}{"extt:$name"} =
	    &make_default_scorer(&make_external_thresholder($external_thresholder));
    }
    if (defined $external_scorer) {
	my $name = $external_scorer;
	$name =~ s/[^\w.]+/-/g;
	$cmpt->{decision_methods}{"exts:$name"} =
	    &make_external_scorer($external_scorer);
    }

    foreach my $t (@threshold_args) {
	if ($t =~ /^t?(\d+(\.\d+)?)$/) {
	    $t = $1;
	    die "threshold [$t] out of range ([0-1]).\n" if ($t < 0 or $t > 1);
	    $cmpt->{decision_methods}{"t$t"} =
		&make_default_scorer(&make_global_thresholder($t));
	} elsif ($t eq "maxmarg") {
	    $cmpt->{decision_methods}{"maxmarg"} =
		&make_default_scorer(&make_nugget_thresholder(\&maxmarg_learner));
	} elsif ($t eq "quartz") {
	    $cmpt->{decision_methods}{"quartz"} =
		&make_default_scorer(&make_nugget_thresholder(\&quartz_learner));
	} elsif ($t eq "logit") {
	    $cmpt->{decision_methods}{"logit"} =
		&make_default_scorer(&make_nugget_thresholder(\&logit_learner));
	} elsif ($t eq "trog") {
	    $cmpt->{decision_methods}{"trog"} =
		&make_default_scorer(&make_nugget_thresholder(\&trog_learner));
	} elsif ($t eq "art") {
	    $cmpt->{decision_methods}{"art"} =
		&make_default_scorer(&make_external_thresholder("$ENV{NUGGETEER_HOME}/src/bayes-threshold/ruby/bin/simple_gl_qid_nid_selection_thresholds.rb"));
	} elsif ($t eq "arj") {
	    $cmpt->{decision_methods}{"arj"} =
		&make_external_scorer("$ENV{NUGGETEER_HOME}/src/bayes-threshold/ruby/bin/nid_infer_online.rb");
	} else {
	    die "unrecognized threshold method [$t]\n";
	}
    }

    if (scalar(keys %{$cmpt->{decision_methods}}) > 1) {
	foreach my $method (keys %{$cmpt->{decision_methods}}) {
	    my $filename = $outputfile.".".$method;
	    my $fh = FileHandle->new($filename, "w");
	    die "cannot write output file [$filename]: $!\n" unless $fh;
	    $cmpt->{output_handles}{$method} = $fh;
	}
    } else {
	foreach my $method (keys %{$cmpt->{decision_methods}}) {
	    my $fh = FileHandle->new(">$outputfile");  #intentionally letting FileHandle interpret >- as stdout instead of forcing outputfilename as above.
	    die "cannot write output file [$outputfile]: $!\n" unless $fh;
	    $cmpt->{output_handles}{$method} = $fh;
	}
    }
}

#>


#< make_default_scorer

sub make_default_scorer {
    my $thresholder = shift;
    return sub {
	# Which @_ is this, my enclosing scope's or mine? Mine, thank W-ll.
	my ($task, $jdb, $opt, $ofh, $knowledge, $qid, $scores, $responses) =@_;
	#< learn
	$knowledge ||= {_tk=>{}};
	if ($task eq learn) {
	    $knowledge->{_tk} = &${thresholder}($task, $jdb, $opt, $ofh, $knowledge->{_tk});
	    return $knowledge;
	}
	#>
	#< judge
	elsif ($task eq judge) {
	    my %judgements = ();
	    foreach my $rid (keys %$scores) {
		my $rscores = $scores->{$rid};
		my $response = $responses->{$rid};
		$judgements{$rid} = {};
		foreach my $nid (keys %$rscores) {
		    my $score = $rscores->{$nid};
		    my $j = &${thresholder}($task, $jdb, $opt, $ofh,
					    $knowledge->{_tk}, $qid, $nid, $rid,
					    $score, $response);
		    $j = $j?1:0;  # canonicalize
		    $knowledge->{_nuggets}||={};
		    $knowledge->{_nuggets}{$qid}||={};
		    $knowledge->{_nuggets}{$qid}{$nid}++ if $j;
		    $knowledge->{_lengths}{$qid}||={};
		    if (not defined $knowledge->{_lengths}{$qid}{$rid}) {
			my $wordlen = $response; $wordlen =~ s/\s*//g;
			# remove whitespace
			$knowledge->{_lengths}{$qid}{$rid} = length $wordlen;
		    }
		    $judgements{$rid}{$nid} = $j;
		}
	    }
	    return \%judgements;
	}
	#>
	#< qresult
	elsif ($task eq qresult) {
	    my $kb = $knowledge->{_nuggets};
	    confess "missing _nuggets field in knowledge: ".Dumper($knowledge)
		unless defined $kb and "HASH" eq ref $kb;

	    my $vital_match = 0;
	    my $total_match = 0;
	    foreach my $nid (keys %{$kb->{$qid}}) {
		$vital_match += $jdb->{_vitoks}{$qid}{$nid};
		$total_match++;
	    }
	    my $vital_poss = 0;
	    my $okay_poss = 0;
	    foreach my $v (values %{$jdb->{_vitoks}{$qid}}) {
		$vital_poss += $v;
		$okay_poss++ if $v == 0;
	    }

	    my $okay_match = $total_match - $vital_match;
	    my $total_poss = $vital_poss;
	    die "qid $qid has no vital or non-zero score nuggets?\n"
		unless $vital_poss;


	    my $recall = $vital_match / $vital_poss;

	    my $allowance = 100 * $total_match;
	    my $length = 0;
            for my $rid (keys %{$knowledge->{_lengths}{$qid}}) {
		$length += $knowledge->{_lengths}{$qid}{$rid};
            }
	    my $precision = ($length < $allowance) ? 1
		: (($length) ? (1 - ($length - $allowance) / $length) : 0 );

	    my $fmeasure = fmeas($precision, $recall, $jdb->{_settings}{beta});

	    return { vital_match => $vital_match,
		     vital_poss => $vital_poss,
		     okay_match => $okay_match,
		     okay_poss => $okay_poss,
		     total_match => $total_match,
		     total_poss => $total_poss,
		     length => $length,
		     allowance => $allowance,
		     recall => $recall,
		     precision => $precision,
		     fmeasure => $fmeasure,
	    };
	} else {
	    die "unrecognized task [$task]\n";
	}
	#>
    };
}

#>
#< make_external_scorer

our $INFERIOR_ERRFILE;
BEGIN {
    $INFERIOR_ERRFILE = qq(/tmp/nuggeteer-extern-err-$$);
    unlink $INFERIOR_ERRFILE;
    $SIG{PIPE} = \&sig_pipe;
}
sub sig_pipe {
    open(ERRFILE, $INFERIOR_ERRFILE) or 
	die "caught sigpipe (a program that nuggeteer called died unexpectedly) [$@] but cannot read inferior errfile [$INFERIOR_ERRFILE]: $!\n";
    my $error = join("", <ERRFILE>);
    die "caught sigpipe (a program that nuggeteer called died) with message:\n"
	.$error."\nthat message also lives in $INFERIOR_ERRFILE\n";
}
sub make_external_scorer {
    my $command = shift;
    return sub {
	my ($task, $jdb, $opt, $ofh, $knowledge, $qid, $scores, $responses) =@_;
	#< learn knowledge = { qid => { nid => threshold } }
	if ($task eq learn) {
	    $knowledge ||= {};

	    my $training_file = "/tmp/nuggeteer-extern-score-training-$$";
	    write_training_file($jdb, $training_file);
	    my $ans_key_spec = join(",", @{$jdb->{_settings}{answerkeys}});
	    my $runcmd = qq($command --quiet --answer_keys $ans_key_spec $training_file 2>> $INFERIOR_ERRFILE);
	    print $ofh "#EXTERN_COMMAND: $runcmd\n" unless $opt->{brief};

	    use IPC::Open2;
	    my ($kid_in, $kid_out);
	    my $pid = open2($kid_out, $kid_in, $runcmd);
	    die "couldn't run command [$runcmd]: $@\n`cat $INFERIOR_ERRFILE`" unless $pid;

	    $knowledge->{_kid_in} = $kid_in;
	    $knowledge->{_kid_out} = $kid_out;
	    # Set autoflush on $kid_in
	    select $kid_in; $| = 1; select STDOUT;
	    # Is there a reason for:
	    # $| = 1, select $_ for select $kid_in;
	    return $knowledge;
	}
	#>
	#< judge a question
	elsif ($task eq judge) {
	    my $kid_in  = $knowledge->{_kid_in};
	    my $kid_out = $knowledge->{_kid_out};
	    confess "external command unprepared\n" .
                Dumper($kid_in, $kid_out)
		unless $kid_in and $kid_out;

	    my %judgements = ();
	    foreach my $rid (keys %$scores) {
		my $rscores = $scores->{$rid};
		my $response = $responses->{$rid};
		$judgements{$rid} = {};


		my @nidscores = map { qq($_=>$rscores->{$_}); } keys %$rscores;
		my $score_input = join(",", @nidscores);
		send_request($kid_in,
			     join("\t", "JUDGE_REQUEST",
				  $qid, $score_input, $response) . "\n",
			     $opt);

		# external magic happens.

		my $judgement = receive_response($kid_out, $opt);
		$judgement =~ s/^NUGGET_PROBS_LIST\s*//;
		foreach my $nidjudgement (split ', ', $judgement) {
		    my ($nid, $j) = split /\s*=>\s*/, $nidjudgement;
		    # TODO how do we deal with probabilistic judgements?
		    # do $j=$j?1:0;? leave them be?
		    $judgements{$rid}{$nid} = $j;
		}
	    }
	    return \%judgements;
	}
	#>
	#< qresult
	elsif ($task eq qresult) {
	    my $kid_in  = $knowledge->{_kid_in};
	    my $kid_out = $knowledge->{_kid_out};
	    my $kid_err = 1;
#	    my $kid_err = $knowledge->{_kid_err};
	    die "external command unprepared\n" .
                Dumper($kid_in, $kid_out, $kid_err)
		unless $kid_in and $kid_out and $kid_err;

	    send_request($kid_in, "QRESULT_REQUEST\t$qid\n", $opt);

	    # external magic happens

	    my $qresult = receive_response($kid_out, $opt);
	    $qresult =~ s/^QRESULT_RESPONSE\s*//;
	    my ($vital_match, $vital_poss, $okay_match, $okay_poss,
		$total_match, $total_poss, $length, $recall, $allowance,
		$precision, $fmeasure) = split /\t/, $qresult;

	    return { vital_match => $vital_match,
		     vital_poss => $vital_poss,
		     okay_match => $okay_match,
		     okay_poss => $okay_poss,
		     total_match => $total_match,
		     total_poss => $total_poss,
		     length => $length,
		     allowance => $allowance,
		     recall => $recall,
		     precision => $precision,
		     fmeasure => $fmeasure,
	    };
	} else {
	    die "unexpected task [$task]\n";
	}
	#>
    }
}

sub send_request {
    my ($handle, $request, $opt) = @_;
    print "Issuing request $request" if $opt->{debug};
    print $handle $request;
}

sub receive_response {
    my ($handle, $opt) = @_;
    my $response = <$handle>;
    print "Got response $response" if $opt->{debug};
    chomp $response;
    return $response;
}
#>


# Thresholders take the score for a particular response-nugget
# decision and return true or false.  Of course this doesn't have to
# be a thresholding method only, but it's what this form has been used
# for so far.  It assumes, unlike the scorer, that decisions about
# each response-nugget pair are unrelated to each-other.

#< make_external_thresholder(command)

sub make_external_thresholder {
    my $command = shift;
    return sub {
	my ($task, $jdb, $opt, $ofh, $knowledge, $qid, $nid, $rid, $score, $response) = @_;
	#< learn knowledge = { qid => { nid => threshold } }
	if ($task eq learn) {
	    $knowledge ||= {};

	    my $training_file = "/tmp/nuggeteer-externthresh-$$";
	    write_training_file($jdb, $training_file);
	    print $ofh "#EXTERN_COMMAND: $command --quiet < $training_file\n"
		unless $opt->{brief};
	    my $runcmd = qq($command --quiet < $training_file 2>\&1);
	    open(EXTERN, "$runcmd |")
		or die "cannot run [$runcmd]: $!$@\n";
	    while (<EXTERN>) {
		next if /^\#/ or !/\S/;
		print $ofh "\#EXTERN: $_" unless $opt->{brief};
		chomp;
		my ($level, $qid, $nid, $t) = split /\t/;
		if ($level eq "GLOBAL" && defined $qid) {
		    $knowledge->{_} = $qid;
		} elsif ($level eq "QUESTION" && defined $qid && defined $nid) {
		    $knowledge->{$qid}||={};
		    $knowledge->{$qid}{_} = $nid;
		} elsif ($level eq "NUGGET" && defined $qid && defined $nid
			 && defined $t) {
		    $knowledge->{$qid}||={};
		    $knowledge->{$qid}{$nid}=$t;
		} else {
		    warn "expected [level \\t qid \\t nid \\t threshold] in external output but found:$.: [$_]\n";
		}
	    }

	    return $knowledge;
	}
	#>
	#< judge
	elsif ($task eq judge) {
	    if (defined $knowledge->{$qid}) {
		if (defined $knowledge->{$qid}{$nid}) {
		    return $score >= $knowledge->{$qid}{$nid};
		} elsif (defined $knowledge->{$qid}{_}) {
		    return $score >= $knowledge->{$qid}{_};
		} elsif (defined $knowledge->{_}) {
		    return $score > $knowledge->{_};
		} else {
		    # TODO I am not sure that this is the right error message
		    die "unexpected knowledge format qid $qid: ".Dumper($knowledge->{$qid});
		}
	    } elsif (defined $knowledge->{_}) {
		return $score > $knowledge->{_};
	    } else {
		die "unexpected knowledge format: ".Dumper($knowledge);
	    }
	}
	#>
	#< unsupported
	else {
	    confess "task $task is unsupported\n";
	}
	#>
    };
}

#>
#< make_global_thresholder

sub make_global_thresholder {
    my $threshold = shift;
    return sub {
	my ($task, $jdb, $opt, $ofh, $knowledge, $qid, $nid, $rid, $score, $response) = @_;
	$knowledge ||= {};
	return $knowledge if $task eq learn;
	return $score >= $threshold if $task eq judge;
	confess "I don't support the [$task] operation";
    };
}

#>
#< make_nugget_thresholder
sub make_nugget_thresholder {
    my $learner=shift; # need to be able to call:
    # threshold = learner(\@positives, \@negatives, $brief)
    return sub {
	my ($task, $jdb, $opt, $ofh, $knowledge, $qid, $nid, $rid, $score, $response) = @_;
	#learn knowledge = { qid => { nid => threshold } }
	if ($task eq learn) {
	    $knowledge ||= {};

	    my $nthr = 0;
	    my $sthr = 0;

	    foreach my $qid (@{$jdb->{_qids}}) {
		my $qthr = {};
		$knowledge->{$qid} = $qthr;

		foreach my $nid (keys %{$jdb->{_nuggets}{$qid}}) {
		    my @negatives = reverse sort {$a <=> $b}
		    values %{$jdb->{_known_neg}{$qid}{$nid}};
		    my @positives = sort {$a <=> $b}
		    values %{$jdb->{_known_pos}{$qid}{$nid}};

		    my $threshold = &$learner(\@positives, \@negatives,
					      $opt->{brief}?$ofh:undef,
					      $opt, $qid, $nid);
		    $qthr->{$nid} = $threshold;
		    $sthr += $threshold;
		    $nthr ++;
		}
            }
	    $knowledge->{_} = $sthr / $nthr;
	    return $knowledge;
	}
	elsif ($task eq judge) {
	    if (defined $knowledge->{$qid} and
		defined $knowledge->{$qid}{$nid}) {
		return $score >= $knowledge->{$qid}{$nid};
	    } else {
		return $score >= $knowledge->{_};
	    }
	}
	else {
	    confess "I don't support the [$task] operation";
	}
    };
}
#>



# To help create thresholders, we have various learners, which support
# just the learn operation, and just on individual nuggets.

#< "maxmarg" score-maximizing margin learner
sub maxmarg_learner {
    my ($p, $n, $ofh, $opt, $qid, $nid) = @_;
    my @positives = @$p;
    my @negatives = @$n;

    if (@negatives) {
	if (@positives) {
	    my $minp = $positives[0];
	    my $maxn = $negatives[0];
	    my $p = 0;
	    my $n = 0;
	    my $min = scalar(@positives) + scalar(@negatives);
	    my @mint = ();
	    while ($p < @positives and $n < @negatives) {
		my $imin = (scalar(@positives) - $p) +
		    (scalar(@negatives) - $n);
		my $pi = $positives[$p];
		my $ni = $negatives[$n];
		my $ti = ($pi+$ni) / 2;
		if ($imin < $min) {
		    @mint = ($ti);
		    $min = $imin;
		} elsif ($imin == $min) {
		    push @mint, $ti;
		}

		if ($pi < $ni) { $p++; }
		else { $n++; }
	    }
	    if (scalar(@positives) < $min) {
		@mint = ($maxn);
	    }

	    print $ofh "\#MAXMARG $qid;$nid = ".join(", ", @mint)."\n" if $ofh;
	    return shift @mint;

	} else { # negs but no pos
	    print $ofh "\#MAXMARG $qid;$nid = top negative\n" if $ofh;
	    return shift @negatives;
	}
    } else {
	if (@positives) {
	    print $ofh "\#MAXMARG $qid;$nid = smallest positive\n" if $ofh;
	    return shift @positives;
	} else {
	    warn "There are really no examples for $qid;$nid?\n"
		if $nid and $opt->{warnings};
	    print $ofh "\#MAXMARG $qid;$nid = 1 (no examples)\n" if $ofh;
	    return 1; # no info at all, don't allow guessing
	}
    }
}
#>
#< "quartz" quartile margin learner
sub quartz_learner {
    my ($p, $n, $ofh, $opt, $qid, $nid) = @_;
    my @positives = @$p;
    my @negatives = @$n;

    my @nzneg = grep { $_ } @negatives;
    my @nzpos = grep { $_ } @positives;
    my $qn = $#nzneg * .75;
    $qn = ($qn < 0)? 0 : $nzneg[int($qn)];
    die "qn".Dumper(@negatives) unless defined $qn;
    my $qp = $#nzpos * .25;
    $qp = ($qp < 0)? 1 : $nzpos[int($qp)];
    die "qp".Dumper(@positives) unless defined $qp;
    my $quartz = ($qn + $qp) / 2;
    print $ofh "\#QUARTZ $qid;$nid = $quartz\n" if $ofh;
    return $quartz;
}
#>
#< "logit" broken logistic regression learner
sub logit_learner {
    my ($p, $n, $ofh, $opt, $qid, $nid) = @_;

    my $sump = 0;
    map { $sump += $_ } @$p;
    map { $sump -= $_ } @$n;
    my $logitp = 1 / (1 + exp(-$sump));
    print $ofh "\#LOGIT $qid;$nid = $logitp\n" if $ofh;
    return $logitp;
}
#>
#< "trog" trying to combine quartz and logit to ill effect.
sub trog_learner {
    my ($p, $n, $ofh, $opt, $qid, $nid) = @_;

    my @nzneg = grep { $_ } @$n;
    my @nzpos = grep { $_ } @$p;
    my $qn = $#nzneg * .75;
    die "qn".Dumper(@$n) unless defined $qn;
    my $qp = $#nzpos * .25;

    my @trimmed_neg = ();
    @trimmed_neg = @nzneg[0..$qn] unless $qn < 0;
    my @trimmed_pos = ();
    @trimmed_pos = @nzpos[0..$qp] unless $qp < 0;

    my $sump = 0;
    map { $sump += $_ } @trimmed_pos;
    map { $sump -= $_ } @trimmed_neg;
    my $logitp = 1 / (1 + exp(-$sump));
    print $ofh "\#TROG $qid;$nid = $logitp\n" if $ofh;
    return $logitp;
}
#>

#>

#< read_testfile(\%opt, $filename)
sub read_testfile {
    my ($opt, $filename) = @_;

    my $filehandle = FileHandle->new("<".$filename)
	or die "cannot read test file [$filename]: $!\n";

    my $result = { qids => [], # list of qids seen
		   responses => {}, #{qid=>{rid=> normalized response text }}
		   originals => {}, #{qid=>{rid=> original response text }}
		   selftest  => {}, #{qid=>{ nid => { rid => 1 } } }
    };

    #An output file is just like a judgement file, but may omit the judgements.
    # response => "qid runtag rid docid response";
    # question => "Submission for question dddd with items numbered";
    # nuggetsline => "Nuggets assigned to items for question dddd";
    # assignment => "qid runtag rid nid";

    my $expecting = response;
    my $rid = 0;
    my %seen = ();
    my @qids = ();
    my $lastqid;
    while (<$filehandle>) {
	chomp;
	next if /^\#/;
	if ($expecting eq question or $expecting eq response) {
	    my ($qid, $runtag, $docid, $resp);
	    #< Nuggets assigned
	    if (/^Nuggets assigned to items for question ($opt->{qidre}):$/) {
		#< this is really a judgement file, so self-test
		my $qid = $1;
		my $selftest = {};
		$_ = <$filehandle>;
		while (defined $_ and not /^$/) {
		    # judgements end with a blank line or end of file I guess.
		    my ($qid, $runtag, $rid, $nid) = split /\s+/;
		    $selftest->{$nid} ||= {};
		    $selftest->{$nid}{$rid}=1;
		    $_ = <$filehandle>;
		}
		$result->{selftest}{$qid}=$selftest;
		#>
		next;
	    }
	    #>
	    #< Submission for question
	    elsif (/^Submission for question ($opt->{qidre}) with items numbered:$/) {
		push @qids, $1;
		$expecting = response;
		$rid = 0;
		next;
	    }
	    #>
	    #< normal response
	    elsif (/^($opt->{qidre})\t(\S+)\t(\S+)\t([^\t]*)$/) {
		($qid, $runtag, $docid, $resp) = split /\t/;
		$rid++;
		if ($qid ne $lastqid) {
		    push @qids, $qid;
		    $rid=1;
		}
		$lastqid = $qid;
	    }
	    #>
	    #< response in judgement file
	    elsif (/^($opt->{qidre})\s+(\S+)\s+(\d+)\s+(\S+)\s+(.*)$/) {
		($qid, $runtag, $rid, $docid, $resp) = ($1, $2, $3, $4, $5);
	    }
	    #>
	    #< blank line
	    elsif (/^\s*$/) {
		$expecting = question;
		next;
	    }
	    #>
	    #< unrecognized
	    else {
		warn "Unrecognized format in [$filename]:line $.: expected [$expecting] with qid=~/$opt->{qidre}/, but got [$_]" if $opt->{warnings};
		next;
	    }
	    #>
	    #< suspicious response: looks like nugget assignment
	    if ($docid =~ /^\d+$/ and $resp =~ /^\d+$/) {
		warn "Response line looks suspiciously like a nugget assignment in [$filename]:line $.: expected [$expecting] but got [$_]\n" if $opt->{warnings};
	    }
	    #>
	    $opt->{runtag} ||= $runtag;

	    next if $opt->{max_input_rank} and
		(++$seen{$qid}) > $opt->{max_input_rank};

	    $result->{originals}{$qid}||={};
	    $result->{originals}{$qid}{$rid} = $resp;

	    $result->{responses}{$qid}||={};
	    $result->{responses}{$qid}{$rid} = lc(trim($resp));
	} else {
	    die "Should not be expecting [$expecting] in a test file!\n";
	}
    }
    $result->{qids} = \@qids;
    print STDERR "read responses for ".scalar(@qids)." questions \n"
	unless $opt->{quiet};
    return $result;
}
#>
#< print_qresult($qid, \%qresult, $handle, \%opt, \%settings)

sub print_qresult {
    my ($qid, $qresult, $out, $opt, $settings) = @_;

    if ($opt->{output_format} eq "trec.eval") {
	print($out qq(Qid: $qid\t$opt->{runtag}\n));
	print($out qq(\tVital matches:   ).
	      $qresult->{vital_match}."/".$qresult->{vital_poss}."\n");
	print($out qq(\tTotal matches:   ).
	      $qresult->{total_match}."/".$qresult->{total_poss}."\n");
	print($out qq(\tLength of response: $qresult->{length}\n));
	printf($out qq(\tRecall: $opt->{number_format}\n), $qresult->{recall});
	printf($out qq(\tPrecision $opt->{number_format}\n),
	       $qresult->{precision});
	printf($out qq(\tF (with beta=$settings->{beta}): ).
	       $opt->{number_format}."\n", $qresult->{fmeasure});
    } else {
	print $out join("\t", $qid, $opt->{runtag},
			sprintf($opt->{number_format},$qresult->{vital_match}), 
			sprintf($opt->{number_format}, $qresult->{okay_match}),
			$qresult->{length},
			sprintf($opt->{number_format}, $qresult->{recall}),
			sprintf($opt->{number_format}, $qresult->{precision}),
			sprintf($opt->{number_format}, $qresult->{fmeasure}))
	    . $/;
    }
}

#>
#< print_judgement($fh, \%judgement, \%opt, qid, nid, rid, j, score, response)
sub print_judgement {
    my ($fh, $jdb, $opt, $qid, $nid, $rid, $j, $score, $response) = @_;
    if ($rid == 1 and $nid == 1) {
	print $fh qq(\#QUESTION $qid: $jdb->{_questions}{$qid}\n);
    }
    confess "did not get expected values for all arguments\n"#.Dumper(@_)
	unless (defined $fh and "FileHandle" eq ref $fh and
		defined $jdb and "HASH" eq ref $jdb and
		defined $opt and "HASH" eq ref $opt and
		defined $qid and defined $nid and defined $rid and
		defined $j and # ($j eq 1 or $j eq 0) and
		defined $score and 
		defined $response);
    # TODO Why is $score sometimes slightly larger than 1?
    # It has occasionally been off by a small enough amount not to
    # register as larger when printed.
    # print STDERR sprintf("%1.50f", $score) if $score eq "1" and $score != 1;
    $score = 1 if $score >= 1 and $score <= 1 + 1e-10;
    confess "score [$score] for question-nugget-response [$qid]-[$nid]-[$rid] out of range\n" 
	unless ($score >= 0 and $score <= 1);
    print $fh join("\t",
		   "#JUDGEMENT", $qid, $rid, $nid, sprintf($opt->{number_format}, $j), sprintf($opt->{number_format}, $score));
    print $fh "\t".join("\t", $response,
			$j?"ASSIGNED":"NOT", $jdb->{_nuggets}{$qid}{$nid})
	if $opt->{debug};
    print $fh "\n";
}
#>

#< trim(string)
sub trim {
    local $_ = shift;
    $_ = "" unless defined $_;
    s/\s+/ /gs;
    s/^ //;
    s/ $//;
    return $_;
}
#>


##################################

#< TODO: judgement format output
#  if (defined $opt->{review_output}) {
#     open(UNKOUT, ">", $opt->{review_output}) or do {
# 	warn "cannot write [$opt->{review_output}]: $!\n";
# 	undef $opt->{review_output};
#     }
# }
# if ($opt->{review_output} and $unknass) {
#     print UNKOUT "\nNuggets assigned to items for question $lastqid:\n";
#     print UNKOUT $unknass;
#     print UNKOUT qq(\n\n\n);
# }
# close UNKOUT if $opt->{review_output};
#>
#< TODO: self test
#     # %ourass needs to be correct for this to produce useful results.
#     if (/^Nuggets assigned to items for question ([\d\.]+):$/) {
# 	$qid = $1;
# 	$_ = <TESTFILE>;
# 	my $tru = 0;
# 	my $obs = 0;
# 	foreach my $k (keys %ourass) { $obs += keys(%{$ourass{$k}}); }
# 	my $cor = 0;
# 	while (!/^$/) {
# 	    chomp;
# 	    my ($qid, $runtag, $osid, $nid) = split /\s+/;
# 	    $tru++;
# 	    $cor++ if exists $ourass{$osid} and exists $ourass{$osid}{$nid};
# 	    $_ = <TESTFILE>;
# 	}
# 	%ourass = ();
# 	$sid=0;
#         my $recall = ($tru)?$cor/$tru:1;
#         my $precision = ($obs)?$cor/$obs:$recall;
#         print join("\t", "\#EVAL", $qid, $tru, $obs, $cor,
#                    $precision, $recall, fmeas($precision, $recall, 1)).$/;
#     }

#>
#< TODO: known true and false
# while (<TESTFILE>) {
#     if (exists $truth{$qid}{$string}) {
# 	record_known_truth($qid, $runtag, $sid, $string, $docid);
# 	submit_external_judgement_request($qid, $string) if $external_judge;
#     } elsif (exists $false{$qid}{$string}) {
# 	record_known_falsehood($qid, $runtag, $sid, $string, $docid);
# 	submit_external_judgement_request($qid, $string) if $external_judge;
#     } else {
# 	handle_unknown($qid, $runtag, $sid, $string, $docid, $origstring);
# 	# Submits a request intrenally to avoid computing scores twice.
#     }
# }
# print STDERR "\n" unless $opt->{quiet};
# print "#found $ncorrect correct, $nwrong wrong, and $nunknown unknown responses.\n";

#>

#< submit_external_judgement_request*

# sub submit_external_judgement_request {
#     my ($qid, $string) = @_;
#     my %explanations = ();
#     my %scores = score($ndb, $settings, $qid, $string, \%explanations);
#     submit_external_judgement_request_with_scores($qid, $string, \%scores);
# }

# sub submit_external_judgement_request_with_scores {
#     my ($qid, $string, $scoreref) = @_;
#     my %scores = %$scoreref;

#     my $request = "JUDGE_REQUEST\t$qid\t";
#     my $nids_started = 0;
#     foreach my $nid (sort {$scores{$b} <=> $scores{$a}} keys %scores) {
# 	$request .= "," if $nids_started;
# 	$request .= "$nid=>$scores{$nid}";
# 	$nids_started = 1;
#     }
#     $request .= "\t$string\n";

#     print JUDGE $request;
#     print "#$request";
# }

#>
#< record_known_truth
# Reads global %truth
# Writes globals %results, %ourass, %support, $ncorrect, $icorrect
# sub record_known_truth {
#     my ($qid, $runtag, $sid, $string, $docid)= @_;
#     my %nids = @{$truth{$qid}{$string}};
#     foreach my $nid (keys %nids) {
# 	#< record the matched nuggets
# 	my $tdocid = $nids{$nid};
# 	next if exists $support{$qid} and exists $support{$qid}{$nid};

# 	$results{$qid}||={};
# 	$results{$qid}{$nid} = $string;

# 	$ourass{$sid}||={};
# 	$ourass{$sid}{$nid}++;

# 	if ($docid eq $tdocid) {
# 	    # this is still the wrong condition because i only
# 	    # allow one tdocid, which is sort of presumably the
# 	    # latest per string.  I should keep all of them and
# 	    # mark any as supported.

# 	    $support{$qid}||={};
# 	    $support{$qid}{$nid} = $docid;
# 	    $ncorrect++;
# 	}
# 	#>
#     }
#     print STDERR "+" unless $opt->{quiet};
#     $icorrect++;
# }
#>
#< record_known_falsehood
# Writes global $nwrong
# sub record_known_falsehood {
#     print STDERR "-" unless $opt->{quiet};
#     $nwrong++;
#     # do nothing.
# }
#>
#< handle_unknown
# Reads globals %unknown, $opt->{review_output}, $lastqid, $unknass, $lruntag, %ndb, %support
# Writes globals %unknown, %ourass, %results, %support, $unknass, $ncorrect, $icorrect, $numknown
# sub handle_unknown {
#     my ($qid, $runtag, $sid, $string, $docid, $origstring)= @_;
#     die "Unseen response in qid $qid: [$string]\n" if $opt->{ensure_known};
#     $unknown{$qid}||=0;
#     $unknown{$qid}++;
#     #< print this item to the unknowns file

#     #warn "cannot find $qid [$string]\n";
#     if ($opt->{review_output}) {
# 	if ($unknown{$qid} == 1) {
# 	    if ($lastqid) {
# 		print UNKOUT "\nNuggets assigned to items for question $lastqid:\n";
# 		print UNKOUT $unknass;
# 		print UNKOUT qq(\n\n\n);
# 		$unknass = "";
# 	    }
# 	    print UNKOUT "Submission for question $qid with items numbered:\n";
# 	}
# 	print UNKOUT qq($qid $lruntag-UNK $unknown{$qid} $docid $origstring\n);
# 	$lastqid = $qid;
#     }

#     #>
#     my $judgement = judge($ndb, $qid, $string);
#     #< record judgement

#     if ($judgement and @$judgement) {
# 	foreach my $nid (@$judgement) {
# 	    next if exists $support{$qid} and exists $support{$qid}{$nid};

# 	    $ourass{$sid}||={};
# 	    $ourass{$sid}{$nid}++;
# 	    $results{$qid}||={};
# 	    $results{$qid}{$nid} = $docid;
# 	    $support{$qid}||={};
# 	    $support{$qid}{$nid} = $docid;
# 	    $unknass .= "$qid\t$lruntag-UNK\t$unknown{$qid}\t$nid\n";

# 	    $ncorrect++
# 	    }
# 	$icorrect++;
# 	print STDERR "*" unless $opt->{quiet};
#     } else {
# 	$nunknown++;
# 	print STDERR "x" unless $opt->{quiet};
#     }

#     #>
# }
#>



1;
